######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Runner to interact with a HTTP gateway to Khiops"""

import io
import json
import os
import ssl
import tempfile
import uuid
import warnings
from urllib.request import Request, urlopen

import khiops.core.internals.filesystems as fs
from khiops import core as kh
from khiops import get_compatible_khiops_version
from khiops.core.exceptions import KhiopsEnvironmentError
from khiops.core.internals.common import type_error_message
from khiops.core.internals.runner import KhiopsRunner
from khiops.core.internals.task import KhiopsTask


class KhiopsDockerRunner(KhiopsRunner):
    """Implementation of a dockerized remote Khiops runner

    Requires a running docker Khiops instance

    Parameters
    ----------

    url : str
        URL for the Docker Khiops server.
    shared_dir : str
        Location of the shared directory. May be an URL/URI.
    insecure : bool, default ``False``
        If ``True`` the target server an HTTPS URL connection requires a certificate.
    """

    def __init__(self, url, shared_dir, insecure=False):
        # Check the parameters
        if not isinstance(url, str):
            raise TypeError(type_error_message("url", url, str))
        if not isinstance(shared_dir, str):
            raise TypeError(type_error_message("shared_dir", shared_dir, str))
        elif fs.is_local_resource(shared_dir) and not fs.exists(shared_dir):
            raise KhiopsEnvironmentError(
                f"'shared_dir' does not exists. Path: {shared_dir}."
            )

        # Call parent constructor
        super().__init__()

        # Initialize this class specific members
        if not url.endswith("/"):
            url += "/"
        self._url = url
        self._insecure = insecure
        self.shared_dir = shared_dir
        # Define SSL context only for https
        if url.startswith("https"):
            # Compare URL host and certificate contents
            ctx = ssl._create_unverified_context()
            ctx.check_hostname = True
            ctx.verify_mode = ssl.CERT_REQUIRED
            # Ignore SSL certificate errors
            if self._insecure:
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
            self._ctx = ctx
        else:
            self._ctx = None

    def _initialize_khiops_version(self):
        self._khiops_version = get_compatible_khiops_version()

    def _get_khiops_version(self):
        if self._khiops_version is None:
            self._initialize_khiops_version()
        return self._khiops_version

    def _fetch(self, request):
        # HTTPS mode
        if self._ctx:
            with urlopen(request, context=self._ctx) as response_obj:
                response = response_obj.read().decode()
        # HTTP mode
        else:
            with urlopen(request) as response_obj:
                response = response_obj.read().decode()
        return response

    def _create_scenario_file(self, task):
        assert isinstance(task, KhiopsTask)
        return self._create_local_temp_file(f"{task.name}_", "._kh")

    def _run(
        self,
        tool_name,
        scenario_path,
        command_line_options,
        trace,
    ):
        # Check arguments
        if not command_line_options.batch_mode:
            warnings.warn(
                "Ignoring unsupported KhiopsDockerRunner parameter 'batch_mode'",
                stacklevel=3,
            )
        if command_line_options.output_scenario_path:
            output_scenario_dir = fs.get_parent_path(
                command_line_options.output_scenario_path
            )
            if not fs.exists(output_scenario_dir):
                fs.make_dir(output_scenario_dir)
            fs.copy_from_local(command_line_options.output_scenario_path, scenario_path)
        if command_line_options.task_file_path is not None:
            warnings.warn(
                "Ignoring unsupported KhiopsDockerRunner parameter 'task_file_path'"
            )

        # Submit the job  for remote execution
        url = f"{self._url}v1/batch"
        if trace:
            print(f"Docker runner URL: {url}")
        tool_flag = "KHIOPS"
        if tool_name == "khiops_coclustering":
            tool_flag = "KHIOPS_COCLUSTERING"
        shared_scenario_path = self.create_temp_file("scenario", "._kh")
        fs.copy_from_local(shared_scenario_path, scenario_path)
        post_fields = {"tool": tool_flag, "scenario_path": shared_scenario_path}
        request = Request(url, data=bytes(json.dumps(post_fields), encoding="utf-8"))
        json_response = self._fetch(request)
        response = json.loads(json_response)

        # If trace is on: display call arguments
        if trace:
            print(f"Docker runner response: {json_response}")

        # Raise an exception if the job could not be sent
        if "error" in response:
            error_message = (
                f"{tool_name} execution failed with return code "
                f"{response['error']['code']}, "
                f"msg = {response['error']['message']}"
            )
            raise kh.KhiopsRuntimeError(
                f"KhiopsDockerRunner could not send job "
                f"to server. Error:\n{error_message}"
            )

        # Repeatedly query operation status, waiting until it is 'done'
        job_id = response["name"]
        if trace:
            print(f"KhiopsDockerRunner job id: {job_id}")
        while not bool(response["done"]):
            url = f"{self._url}v1/operation/{job_id}:wait?timeout=10s"
            request = Request(url)
            json_response = self._fetch(request)
            response = json.loads(json_response)
            if trace:
                print(f"Docker runner fetch job status: {json_response}")

        # Once done obtain the Khiops return code, stderr
        # The stdout ("output" field) is written to the specified log file
        # Then report (warn or exception) any errors
        return_code = response["response"]["status"]
        stderr = response["response"]["error"]
        with io.StringIO() as log_file_stream:
            log_file_stream.write(response["response"]["output"])
            fs.write(
                command_line_options.log_file_path,
                log_file_stream.getvalue().encode("utf8", errors="replace"),
            )

        # Delete the finished operation from the server
        url = f"{self._url}v1/operation/{job_id}:delete"
        request = Request(url, method="DELETE")
        json_response = self._fetch(request)
        if trace:
            print(f"Docker runner delete job: {json_response}")

        # We return empty stdout because the 'output' field of the JSON response are the
        # contents of the log
        return return_code, "", stderr

    def _create_local_temp_file(self, prefix, suffix):
        """Creates a unique temporary file in the runner's root temporary directory

        Parameters
        ----------
        prefix : str
            Prefix for the file's name.

        suffix : str
            Suffix for the file's name.

        Returns
        -------
        str
            The path of the created file.
        """
        tmp_file_fd, tmp_file_path = tempfile.mkstemp(
            prefix=prefix, suffix=suffix, dir=self.root_temp_dir
        )
        os.close(tmp_file_fd)
        return tmp_file_path

    # Override methods from base class
    def create_temp_file(self, prefix, suffix):
        """Creates a unique temporary file name in the runner's shared directory
        The actual file is not created on disk.

        Parameters
        ----------
        prefix : str
            Prefix for the file's name.

        suffix : str
            Suffix for the file's name.

        Returns
        -------
        str
            The path of the file to be created.
        """
        return fs.get_child_path(self.shared_dir, f"{prefix}{uuid.uuid4()}{suffix}")

    def create_temp_dir(self, prefix):
        """Creates a unique directory in the runner's shared directory

        Parameters
        ----------
        prefix : str
            Prefix for the directory's name.

        Returns
        -------
        str
            The path of the created directory.
        """
        tmp_dir_path = fs.get_child_path(
            self.shared_dir, f"{prefix}{uuid.uuid4()}{os.path.sep}"
        )
        fs.make_dir(tmp_dir_path)
        return tmp_dir_path



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Common utility functions and classes"""
import os
from collections.abc import Iterable, Mapping, Sequence
from urllib.parse import urlparse


class SystemSettings:
    """Khiops system settings

    .. note::
        These settings are not available in the `CommandLineOptions`.
    """

    def __init__(
        self,
        max_cores=None,
        memory_limit_mb=None,
        temp_dir="",
        scenario_prologue="",
    ):
        """See class docstring"""
        self.max_cores = max_cores
        self.memory_limit_mb = memory_limit_mb
        self.temp_dir = temp_dir
        self.scenario_prologue = scenario_prologue
        self.check()

    def __repr__(self):
        return (
            f"System settings: max_cores = {self.max_cores}, "
            f"memory_limit_mb = {self.memory_limit_mb}, "
            f"temp_dir = {self.temp_dir}, "
            f"scenario_prologue = {self.scenario_prologue}"
        )

    def check(self):
        """Checks the types of the system settings

        Raises
        ------
        `TypeError`
            If any of the system settings does not have the proper type.
        `ValueError`
            If ``max_cores`` or ``memory_limit_mb`` are set to negative numbers.
        """
        # Check the field types and ranges where applicable
        if self.max_cores is not None:
            if not isinstance(self.max_cores, int):
                raise TypeError(type_error_message("max_cores", self.max_cores, int))
            if self.max_cores < 0:
                raise ValueError(
                    f"max_cores must be non-negative (it is {self.max_cores})"
                )
        if self.memory_limit_mb is not None:
            if not isinstance(self.memory_limit_mb, int):
                raise TypeError(
                    type_error_message("memory_limit_mb", self.memory_limit_mb, int)
                )
            if self.memory_limit_mb < 0:
                raise ValueError(
                    "memory_limit_mb must be non-negative "
                    f"(it is {self.memory_limit_mb})"
                )
        if not is_string_like(self.temp_dir):
            raise TypeError(
                type_error_message(
                    "temp_dir",
                    self.temp_dir,
                    str,
                    bytes,
                )
            )
        if not is_string_like(self.scenario_prologue):
            raise TypeError(
                type_error_message(
                    "scenario_prologue", self.scenario_prologue, str, bytes
                )
            )


class CommandLineOptions:
    """Khiops command line options

    Attributes
    ----------
    log_file_path : str, default ""
        Path of the log file for the Khiops process (command line option ``-e`` of the
        desktop app). If equal to "" then it writes no log file.
    output_scenario_path : str, default ""
        Path of the output Khiops scenario file (command line option ``-o`` of the
        desktop app). If the empty string is specified no output scenario file is
        generated.
    task_file_path : str, default ""
        Path of the task file for the Khiops process (command line option ``-p`` of the
        desktop app). If equal to "" then it writes no task file.
    batch_mode : bool, default True
        *Deprecated* Will be removed in Khiops 11. If ``True`` activates batch mode
        (command line option ``-b`` of the app).
    """

    def __init__(
        self,
        batch_mode=True,
        log_file_path="",
        task_file_path="",
        output_scenario_path="",
    ):
        """See class docstring"""
        self.batch_mode = batch_mode
        self.log_file_path = log_file_path
        self.task_file_path = task_file_path
        self.output_scenario_path = output_scenario_path
        self.check()

    def __repr__(self):
        def to_str(string):
            if isinstance(string, str):
                repr_str = string
            else:
                assert isinstance(string, bytes)
                repr_str = string.decode("utf8", errors="replace")
            return repr_str

        command_line_options = []
        if self.batch_mode:
            command_line_options += ["-b"]
        if self.output_scenario_path:
            command_line_options += ["-o", to_str(self.output_scenario_path)]
        if self.log_file_path:
            command_line_options += ["-e", to_str(self.log_file_path)]
        if self.task_file_path:
            command_line_options += ["-p", to_str(self.task_file_path)]
        print(command_line_options)
        return "Khiops command line options: " + " ".join(command_line_options)

    def build_command_line_options(self, scenario_path):
        command_line_options = []
        if self.batch_mode:
            command_line_options += ["-b"]
        command_line_options += ["-i", scenario_path]
        if self.output_scenario_path:
            command_line_options += ["-o", self.output_scenario_path]
        if self.log_file_path:
            command_line_options += ["-e", self.log_file_path]
        if self.task_file_path:
            command_line_options += ["-p", self.task_file_path]

        return command_line_options

    def check(self):
        """Checks the types of the command line options

        Raises
        ------
        `TypeError`
            If any of the command line options does not have the proper type.
        """
        if not isinstance(self.batch_mode, bool):
            raise TypeError(type_error_message("batch_mode", self.batch_mode, bool))
        if self.output_scenario_path and not is_string_like(self.output_scenario_path):
            raise TypeError(
                type_error_message(
                    "output_scenario_path", self.output_scenario_path, str, bytes
                )
            )
        if self.log_file_path and not is_string_like(self.log_file_path):
            raise TypeError(
                type_error_message("log_file_path", self.log_file_path, str, bytes)
            )
        if self.task_file_path and not is_string_like(self.task_file_path):
            raise TypeError(
                type_error_message("task_file_path", self.task_file_path, str, bytes)
            )


def create_unambiguous_khiops_path(path):
    """Creates a path that is unambiguous for Khiops

    Khiops needs that a non absolute path starts with "." so that it does not use the
    path of an internally saved state as reference point.

    For example: if we open the data table "/some/path/to/data.txt" and then set the
    results directory simply as "results" the effective location of the results
    directory will be "/some/path/to/results" instead of "$CWD/results". This behavior
    is a feature in the Khiops GUI but it is undesirable when using it as a library.

    This function returns a path so that the library behaves as expected: a path
    relative to the $CWD if it is a non absolute path.
    """
    # Check for string
    if not isinstance(path, (str, bytes)):
        raise TypeError(type_error_message("path", path, str, bytes))

    # Empty path returned as-is
    if not path:
        return path

    # Add a "." to a local path if necessary. It is *not* necessary when:
    # - `path` is an URI
    # - `path` is an absolute path
    # - `path` is a path starting with "."
    dot = "."
    empty = ""
    if isinstance(path, bytes):
        dot = bytes(dot, encoding="ascii")
        empty = bytes(empty, encoding="ascii")
    uri_info = urlparse(path, allow_fragments=False)
    if os.path.isabs(path) or path.startswith(dot) or uri_info.scheme != empty:
        return path
    else:
        return os.path.join(dot, path)


############
# Messages #
############


def type_error_message(variable_name, variable, *target_types):
    """Formats a type error message

    Parameters
    ----------
    variable_name : str
        Name of the variable for whom the type error is signaled.
    variable : any
        Actual variable for whom the type error is signaled.
    target_types : list
        Expected types for ``variable``, either as a type or as a string.

    Returns
    -------
    str
        The type error message.

    """
    assert len(target_types) > 0, "At least one target type must be provided"
    assert all(
        isinstance(target_type, (type, str)) for target_type in target_types
    ), "All target types must be 'type' or 'str'"
    assert isinstance(variable_name, str), "'variable_name' must be 'str'"
    assert len(variable_name) > 0, "'variable_name' should not be empty"

    # Transform to 'type' the string arguments
    typed_target_types = []
    for target_type in target_types:
        if isinstance(target_type, str):
            typed_target_types.append(type(target_type, (), {}))
        else:
            typed_target_types.append(target_type)

    # Build the type error message
    if len(typed_target_types) == 1:
        target_type_str = f"'{typed_target_types[0].__name__}'"
    elif len(typed_target_types) == 2:
        target_type_str = (
            f"either '{typed_target_types[0].__name__}' "
            f"or '{typed_target_types[1].__name__}'"
        )
    else:
        target_types_str = " or ".join(
            f"'{target_type.__name__}'" for target_type in typed_target_types
        )
        target_type_str = f"one of {target_types_str}"

    if len(variable_name.strip().split(" ")) == 1:
        variable_name_str = f"'{variable_name}'"
    else:
        variable_name_str = variable_name

    return (
        f"{variable_name_str} type must be {target_type_str}, "
        f"not '{type(variable).__name__}'"
    )


def removal_message(removed_feature, since, replacement=None):
    """Formats a feature removal message"""
    message = f"'{removed_feature}' removed since {since}. "
    if replacement:
        message += f"Use '{replacement}'."
    else:
        message += "There is no replacement."
    return message


def renaming_message(renamed_feature, new_name, since):
    """Formats a feature renaming message"""
    return f"Ignoring '{renamed_feature}': renamed to '{new_name}' since {since}."


def invalid_keys_message(kwargs):
    """Formats an invalid keyword parameter message"""
    return f"Ignoring invalid parameter(s): {','.join(kwargs.keys())}."


def deprecation_message(
    deprecated_feature, deadline_version, replacement=None, quote=True
):
    """Formats a deprecation message"""
    if quote:
        message = f"'{deprecated_feature}' is deprecated "
    else:
        message = f"{deprecated_feature} is deprecated "
    message += f"and will be removed by version {deadline_version}."
    if replacement is not None:
        if quote:
            message += f" Prefer '{replacement}'."
        else:
            message += f" Prefer {replacement}."
    else:
        message += " There will be no replacement when removed."
    return message


###############
# Type checks #
###############


def is_string_like(test_object):
    """Returns True if an object is a valid Python string or sequence of bytes"""
    return isinstance(test_object, (str, bytes))


def is_list_like(list_like):
    """Returns True if an object is list-like

    An object is ``list-like`` if and only if inherits from `collections.abc.Sequence`
    and it is not `string-like <is_string_like>`
    """
    return isinstance(list_like, Sequence) and not is_string_like(list_like)


def is_dict_like(test_object):
    """Returns True if an object is dict-like

    An object is ``dict-like`` if and only if inherits from the
    `collections.abc.Mapping`.
    """
    return isinstance(test_object, Mapping)


def is_iterable(test_object):
    """Return True if a container object is iterable, but not string-like"""
    return isinstance(test_object, (Sequence, Iterable)) and not is_string_like(
        test_object
    )



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Scikit-Learn interface to the Khiops AutoML suite"""
from khiops.sklearn.estimators import (
    KhiopsClassifier,
    KhiopsCoclustering,
    KhiopsEncoder,
    KhiopsRegressor,
)
from khiops.sklearn.helpers import train_test_split_dataset



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""export_dictionary_as_json task family"""
from khiops.core.internals import task as tm
from khiops.core.internals.types import StringLikeType

# Disable long lines to have readable scenarios
# pylint: disable=line-too-long
TASKS = [
    tm.KhiopsTask(
        "export_dictionary_as_json",
        "khiops",
        "10.0.0",
        [
            ("dictionary_file_path", StringLikeType),
            ("json_dictionary_file_path", StringLikeType),
        ],
        [],
        ["dictionary_file_path", "json_dictionary_file_path"],
        """
        // Dictionary file settings
        ClassManagement.OpenFile
        ClassFileName __dictionary_file_path__
        OK

        // Export dictionary as JSON file
        ClassManagement.ExportAsJSON
        JSONFileName __json_dictionary_file_path__
        OK
        """,
    ),
]



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Classes to access Khiops Coclustering JSON reports

Class Overview
--------------
Below we describe with diagrams the relationships of the classes in this modules. They
are mostly compositions (has-a relations) and we omit native attributes (str, int,
float, etc).

The main class of this module is `CoclusteringResults` and it is largely a composition
of sub-reports objects given by the following structure::

    CoclusteringResults
    |- coclustering_report -> CoclusteringReport

    CoclusteringReport
    |- dimensions -> list of CoclusteringDimension
    |- cells      -> list of CoclusteringCell

    CoclusteringDimension
    |- parts        -> list of CoclusteringDimensionPart
    |- clusters     -> list of CoclusteringCluster
    |- root_cluster -> CoclusteringCluster

    CoclusteringDimensionPartValueGroup
    |- values -> list of CoclusteringDimensionPartValue

    CoclusteringCluster
    |- leaf_part        -> CoclusteringDimensionPart or None
    |- parent_cluster  |
    |- child_cluster1  |-> CoclusteringCluster or None
    |- child_cluster2  |

To have a complete illustration of the access to the information of all classes in this
module look at their ``write_report`` methods which write TSV (tab separated values)
reports.
"""
import io

from khiops.core.exceptions import KhiopsJSONError
from khiops.core.internals.common import type_error_message
from khiops.core.internals.io import (
    KhiopsJSONObject,
    KhiopsOutputWriter,
    flexible_json_load,
)


class CoclusteringResults(KhiopsJSONObject):
    """Main class containing the information of a Khiops Coclustering JSON file

    Parameters
    ----------
    json_data : dict, optional
        Python dictionary representing the data of a Khiops Coclustering JSON report
        file. If not specified it returns an empty instance.

        .. note::
            Prefer either the the `read_coclustering_results_file` function from the
            core API to obtain an instance of this class from a Khiops Coclustering JSON
            file.

    Attributes
    ----------
    tool : str
        Name of the Khiops tool that generated the JSON file.
    version : str
        Version of the Khiops tool that generated the JSON file.
    coclustering_report : `CoclusteringReport`
        Coclustering modeling report.
    """

    def __init__(self, json_data=None):
        """See class docstring"""
        # Initialize super class
        super().__init__(json_data=json_data)

        # Transform to an empty dictionary if json_data is not specified
        if json_data is None:
            json_data = {}

        # Initialize empty report attributes
        self.short_description = json_data.get("shortDescription", "")

        # Initialize the sub-report only if available
        if "coclusteringReport" in json_data:
            self.coclustering_report = CoclusteringReport(
                json_data["coclusteringReport"]
            )
        else:
            self.coclustering_report = None

    def write_report_file(self, report_file_path):
        """Writes a TSV report file with the object's information

        Parameters
        ----------
        report_file_path : str
            Path of the output TSV report file.
        """
        with open(report_file_path, "wb") as report_file:
            writer = self.create_output_file_writer(report_file)
            self.write_report(writer)

    def write_report(self, stream_or_writer):
        """Writes the instance's TSV report to a writer object

        Parameters
        ----------
        stream_or_writer : `io.IOBase` or `.KhiopsOutputWriter`
            Output stream or writer.
        """
        # Check input writer/stream type
        if isinstance(stream_or_writer, io.IOBase):
            writer = self.create_output_file_writer(stream_or_writer)
        elif isinstance(stream_or_writer, KhiopsOutputWriter):
            writer = stream_or_writer
        else:
            raise TypeError(
                type_error_message(
                    "stream_or_writer",
                    stream_or_writer,
                    io.IOBase,
                    KhiopsOutputWriter,
                )
            )

        # Write report
        writer.writeln(f"#Khiops {self.version}")
        writer.writeln(f"Short description\t{self.short_description}")
        if self.coclustering_report is not None:
            self.coclustering_report.write_report(writer)


def read_coclustering_results_file(json_file_path):
    """Reads a Khiops Coclustering JSON report

    Parameters
    ----------
    json_file_path : str
        Path of the JSON report file.

    Returns
    -------
    `.CoclusteringResults`
        An instance of CoclusteringResults containing the report's information.
    """
    return CoclusteringResults(json_data=flexible_json_load(json_file_path))


class CoclusteringReport:
    """Main coclustering report

    A coclustering is an unsupervised data grid equipped with additional structures to
    ease its exploration. In particular, it is a piecewise constant density estimator of
    the data distribution. The additional structures are the following:

    - A cluster hierarchy for each dimension
    - Indicators (such as the interest) for each variable, part and value.

    A coclustering consists of one to many variables (dimensions), where each variable
    is partitioned as:

    - Intervals in the numerical case
    - Individual values or value groups in the categorical case.

    The cross-product of the partitions forms a multivariate partition of cells and
    their frequencies allow to estimate the multivariate density.

    In case of an unsupervised data grid, the cells are described by their index on the
    variable partitions, together with their frequencies.

    Parameters
    ----------
    json_data : dict, optional
        JSON data of the ``coclusteringReport`` field of a Khiops Coclustering JSON
        report file. If not specified it returns an empty instance.

    Attributes
    ----------
    instance_number : int
        Number of individuals in the learning data table.
    cell_number : int
        Number of coclustering cells.
    null_cost : float
        Cost of the null model.
    level : float
        Measure between 0 and 1 measuring the information gain over the null model.
    initial_dimension_number : int
        Initial number of dimensions. The number of dimensions (``len(dimensions)``) may
        be less than this quantity after a simplification (see
        `~.api.simplify_coclustering`).
    frequency_variable : str
        Name of the variable to be aggregated in the cells. By default is the number of
        individuals.
    dictionary : str
        Name dictionary from which the model was learned.
    database : str
        Path of the main training data table file.
    sample_percentage : float
        Percentage of instances used in training.
    sampling_mode : "Include sample" or "Exclude samples"
        Sampling mode used to split the train and datasets.
    selection_variable : str
        Variable used to select instances for training.
    selection_value : str
        Value of ``selection_variable`` to select instances for training.
    dimensions : list of `CoclusteringDimension`
        Coclustering dimensions (variable).
    cells : list of `CoclusteringCell`
        Coclustering cells.
    """

    def __init__(self, json_data=None):
        """See class docstring"""
        # Check the type of json_data
        if json_data is not None and not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Transform to an empty dictionary if json_data is not specified
        if json_data is None:
            json_data = {}
        # Otherwise raise an exception if the summary is not found
        elif "summary" not in json_data:
            raise KhiopsJSONError("'summary' key not found in coclustering report")

        # Initialize summary fields
        json_summary = json_data.get("summary", {})
        self.instance_number = json_summary.get("instances", 0)
        self.cell_number = json_summary.get("cells", 0)
        self.null_cost = json_summary.get("nullCost", 0)
        self.cost = json_summary.get("cost", 0)
        self.level = json_summary.get("level", 0)
        self.initial_dimension_number = json_summary.get("initialDimensions", 0)
        self.frequency_variable = json_summary.get("frequencyVariable", "")
        self.dictionary = json_summary.get("dictionary", "")
        self.database = json_summary.get("database", "")
        self.sample_percentage = json_summary.get("samplePercentage", 0.0)
        self.sampling_mode = json_summary.get("samplingMode", "")
        self.selection_variable = json_summary.get("selectionVariable")
        self.selection_value = json_summary.get("selectionValue")

        # Create dimensions and initialize their summaries
        self.dimensions = []
        self._dimensions_by_name = {}
        if "dimensionSummaries" in json_data:
            for json_dimension_summary in json_data["dimensionSummaries"]:
                dimension = CoclusteringDimension().init_summary(json_dimension_summary)
                self.dimensions.append(dimension)
                self._dimensions_by_name[dimension.name] = dimension

        # Initialize dimensions' partitions
        if "dimensionPartitions" in json_data:
            json_dimension_partitions = json_data["dimensionPartitions"]
            if len(self.dimensions) != len(json_dimension_partitions):
                raise KhiopsJSONError(
                    "'dimensionPartitions' list has length "
                    f"{len(json_dimension_partitions)} instead of "
                    f"{len(self.dimensions)}"
                )
            for i, json_dimension_partition in enumerate(json_dimension_partitions):
                dimension = self.dimensions[i]
                dimension.init_partition(json_dimension_partition)

        # Initialize dimensions' hierarchies
        if "dimensionHierarchies" in json_data:
            json_dimension_hierarchies = json_data["dimensionHierarchies"]
            if len(self.dimensions) != len(json_dimension_hierarchies):
                raise KhiopsJSONError(
                    "'dimensionHierarchies' list has length "
                    f"{len(json_dimension_hierarchies)} instead of "
                    f"{len(self.dimensions)}"
                )
            for i, json_dimension_hierarchy in enumerate(json_dimension_hierarchies):
                dimension = self.dimensions[i]
                dimension.init_hierarchy(json_dimension_hierarchy)

        # Initialize cells
        self.cells = []
        if "cellPartIndexes" in json_data:
            # Check minimum consistency of input data
            if "cellFrequencies" not in json_data:
                raise KhiopsJSONError(
                    "'cellFrequencies' key not found " "but 'cellPartIndexes' found."
                )
            json_cell_frequencies = json_data["cellFrequencies"]
            json_cell_part_indexes = json_data["cellPartIndexes"]
            if len(json_cell_part_indexes) != len(json_cell_frequencies):
                raise KhiopsJSONError(
                    "'cellPartIndexes' length is different from "
                    f"that of 'cellFrequencies': {len(json_cell_part_indexes)} != "
                    f"{len(json_cell_frequencies)}"
                )
            # Create and initialize all cells
            for i, part_indexes in enumerate(json_cell_part_indexes):
                # Initialize cell
                cell = CoclusteringCell()
                cell.frequency = json_cell_frequencies[i]

                # Initialize cell parts:
                # Retrieve part from its index in the partition, per dimension
                for j, index in enumerate(part_indexes):
                    dimension = self.dimensions[j]
                    part = dimension.parts[index]
                    cell.parts.append(part)

                # Add to the cell list
                self.cells.append(cell)

    def get_dimension_names(self):
        """Returns the names of the available dimensions

        Returns
        -------
        list of str
            The names of the available dimensions.
        """
        return [dimension.name for dimension in self.dimensions]

    def get_dimension(self, dimension_name):
        """Returns the specified dimension

        Parameters
        ----------
        dimension_name : str
            Name of the dimension (variable).

        Returns
        -------
        `CoclusteringDimension`
            The specified dimension.

        Raises
        ------
        `KeyError`
            If no dimension with the specified names exist.
        """
        return self._dimensions_by_name[dimension_name]

    def write_report(self, writer):
        """Writes the instance's TSV report to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output stream or writer.
        """
        # Write each section
        self.write_dimensions(writer)
        self.write_coclustering_stats(writer)
        self.write_bounds(writer)
        self.write_hierarchies(writer)
        self.write_compositions(writer)
        self.write_cells(writer)
        self.write_annotations(writer)

    def write_dimensions(self, writer):
        """Writes the "dimensions" section of the TSV report to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.writeln(f"Dimensions\t{len(self.dimensions)}")
        for i, cc_dimension in enumerate(self.dimensions):
            if i == 0:
                cc_dimension.write_dimension_header_line(writer)
            cc_dimension.write_dimension_line(writer)

    def write_coclustering_stats(self, writer):
        """Writes the "stats" section of the TSV report to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.writeln("")
        writer.writeln("Coclustering stats")
        writer.writeln(f"Instances\t{self.instance_number}")
        writer.writeln(f"Cells\t{self.cell_number}")
        writer.writeln(f"Null cost\t{self.null_cost}")
        writer.writeln(f"Cost\t{self.cost}")
        writer.writeln(f"Level\t{self.level}")
        writer.writeln(f"Initial dimensions\t{self.initial_dimension_number}")
        writer.writeln(f"Frequency variable\t{self.frequency_variable}")
        writer.writeln(f"Dictionary\t{self.dictionary}")
        writer.writeln(f"Database\t{self.database}")
        writer.writeln(f"Sample percentage\t{self.sample_percentage}")
        writer.writeln(f"Sampling mode\t{self.sampling_mode}")
        writer.writeln(f"Selection variable\t{self.selection_variable}")
        writer.writeln(f"Selection value\t{self.selection_value}")

    def write_bounds(self, writer):
        """Writes the "bounds" section of the TSV report to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Compute number of numerical dimensions
        numerical_dimension_number = 0
        for cc_dimension in self.dimensions:
            if cc_dimension.type == "Numerical":
                numerical_dimension_number += 1

        # Write "Bounds" section only in case of numerical dimensions
        if numerical_dimension_number > 0:
            writer.writeln("")
            writer.writeln("Bounds")
            writer.writeln("Name\tMin\tMax")
            for cc_dimension in self.dimensions:
                if cc_dimension.type == "Numerical":
                    writer.write(f"{cc_dimension.name}\t")
                    writer.write(f"{cc_dimension.min}\t")
                    writer.writeln(str(cc_dimension.max))

    def write_hierarchies(self, writer):
        """Writes the dimension reports' "hierarchy" sections to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        for cc_dimension in self.dimensions:
            cc_dimension.write_hierarchy(writer)

    def write_compositions(self, writer):
        """Writes the dimensions' "composition" sections to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        for cc_dimension in self.dimensions:
            cc_dimension.write_composition(writer)

    def write_cells(self, writer):
        """Writes the "cells" section of the TSV report to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Write header
        writer.writeln("")
        writer.writeln("Cells")

        # Write table header
        for cc_dimension in self.dimensions:
            writer.write(f"{cc_dimension.name}\t")
        writer.writeln("Frequency")

        # Write cell report lines
        for cc_cell in self.cells:
            cc_cell.write_line(writer)

    def write_annotations(self, writer):
        """Writes the dimensions' "annotation" sections to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Decide whether annotation sections need to be reported
        need_report = False
        for cc_dimension in self.dimensions:
            need_report = need_report or cc_dimension.needs_annotation_report()

        # Write "Annotation" sections for each dimension
        if need_report:
            for cc_dimension in self.dimensions:
                cc_dimension.write_annotation(writer)
        else:
            writer.writeln("")


class CoclusteringDimension:
    """A coclustering dimension (variable)

    A coclustering dimension is a hierarchical clustering of an input variable. The
    leaves of this hierarchy are linked to an element of a partition of the input
    variable. Leaf clusters have variable parts as their children.

    It only has a no-parameter constructor.

    .. note::
        The instance information is initialized with the `init_summary`,
        `init_partition` and `init_hierarchy` methods. Its owner object (class
        `CoclusteringReport`) uses the information found in the fields
        ``dimensionSummaries``, ``dimensionPartitions`` and ``dimensionHierarchies`` to
        coherently initialize the all dimensions with these methods.

    Attributes
    ----------
    name : str
        Name of the variable associated to this dimension.
    type : "Numerical" or "Categorical"
        Dimension type.
    part_number : int
        Number of parts of the variable associated to this dimension.
    initial_part_number : int
        Number of initial parts. Note that ``part_number`` <= ``initial_part_number``
        after a coclustering simplification (see `~.api.simplify_coclustering`).
    value_number : int
        Number of values of the dimension's variable.
    interest : float
        Interest of the dimension with respect to the other coclustering dimensions.
    description : str
        Description of the dimension/variable.
    min : float
        Minimum value of a numerical dimension/variable.
    max : float
        Maximum value of a numerical dimension/variable.
    parts : list of `CoclusteringDimensionPart`
        Partition of this dimension.
    clusters : list of `CoclusteringCluster`
        Clusters of this dimension's hierarchy. Note that includes intermediary
        clusters.
    root_cluster : `CoclusteringCluster`
        Root cluster of the hierarchy.
    """

    def __init__(self):
        """See class docstring"""
        # Summary attributes
        self.name = ""
        self.type = ""
        self.part_number = 0
        self.initial_part_number = 0
        self.value_number = 0
        self.interest = 0
        self.description = ""

        # Min & Max attributes: Numerical dimension only
        self.min = None
        self.max = None

        # List of parts: intervals or value groups according to dimension type
        self.parts = []

        # Default group attribute only for categorical dimensions
        self.default_group = None

        # List of the hierarchy clusters, ranging from the root to the leaves
        self.root_cluster = None
        self.clusters = []

        # Parts internal dictionary
        self._parts_by_name = {}

        # Clusters internal dictionary
        self._clusters_by_name = {}

    def init_summary(self, json_data=None):
        """Initializes the summary attributes from a Python JSON object

        Parameters
        ----------
        json_data : dict, optional
            Dictionary representing the data of an element of the list found at the
            ``dimensionSummaries`` field of a Khiops Coclustering JSON report file. If
            not specified it leaves the object as-is.

        Returns
        -------
        self
            A reference to the caller instance.
        """
        # Do nothing if json_data is not specified
        if json_data is None:
            return self
        # Otherwise check the type of json_data
        elif not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Initialize the summary fields
        self.name = json_data.get("name", "")
        self.type = json_data.get("type", "")
        self.part_number = json_data.get("parts", 0)
        self.initial_part_number = json_data.get("initialParts", 0)
        self.value_number = json_data.get("values", 0)
        self.interest = json_data.get("interest", 0)
        self.description = json_data.get("description", "")
        self.min = json_data.get("min")
        self.max = json_data.get("max")

        return self

    def init_partition(self, json_data=None):
        """Initializes the partition attributes from a Python JSON object

        Parameters
        ----------
        json_data : dict, optional
            Python dictionary representing the data of an element of the list found at
            the ``dimensionPartitions`` field of a Khiops Coclustering JSON report file.
            If not specified it leaves the object as-is.

        Returns
        -------
        self
            A reference to the caller instance.
        """
        # Do nothing if json_data is not specified
        if json_data is None:
            return self
        # Otherwise check the type of json_data
        elif not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Check minimum required attributes
        if self.name != json_data.get("name"):
            raise KhiopsJSONError(
                f"""'name' field value is '{json_data.get("name")}' """
                f"it should be '{self.name}'"
            )
        if self.type != json_data.get("type"):
            raise KhiopsJSONError(
                f"""'name' field value is '{json_data.get("type")}' """
                f"it should be '{self.type}'"
            )

        # Initialize interval partitions
        if self.type == "Numerical":
            # Check the "intervals" field
            if "intervals" not in json_data:
                raise KhiopsJSONError(
                    "'intervals' key not found in numerical partition"
                )
            elif not isinstance(json_data["intervals"], list):
                raise KhiopsJSONError("'intervals' key must be a list")
            elif len(json_data["intervals"]) < 2:
                raise KhiopsJSONError("'intervals' key must have length at least 2")

            # Initialize the intervals data
            for json_interval in json_data["intervals"]:
                interval = CoclusteringDimensionPartInterval(json_interval)
                self.parts.append(interval)
                self._parts_by_name[interval.cluster_name] = interval

            # Initialize open interval flags
            first_interval = self.parts[0]
            if first_interval.is_missing:
                first_interval = self.parts[1]
            first_interval.is_left_open = True
            last_interval = self.parts[-1]
            last_interval.is_right_open = True

        # Initialize value groups partitions
        elif self.type == "Categorical":
            # Initialize regular value groups
            if "valueGroups" not in json_data:
                raise KhiopsJSONError("'valueGroups' key not found")
            for json_value_group in json_data["valueGroups"]:
                value_group = CoclusteringDimensionPartValueGroup(json_value_group)
                self.parts.append(value_group)
                self._parts_by_name[value_group.cluster_name] = value_group

            # Initialize default group
            # The default group contains all the values not specified in
            # any part of the partition and all the unknown values
            default_group_index = json_data.get("defaultGroupIndex")
            self.default_group = self.parts[default_group_index]
            self.default_group.is_default_part = True

        return self

    def init_hierarchy(self, json_data):
        """Initializes the hierarchy attributes from a Python JSON object

        Parameters
        ----------
        json_data : dict, optional
            Python dictionary representing the data of an element of the list found at
            the ``dimensionHierarchies`` field of a Khiops Coclustering JSON report
            file. If not specified it leaves the object as-is.

        Returns
        -------
        self
            A reference to the caller instance.
        """
        # Do nothing if json_data is not specified
        if json_data is None:
            return self
        # Otherwise check the type of json_data
        elif not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Check minimum required attributes
        if self.name != json_data.get("name"):
            raise KhiopsJSONError(
                f"""'name' field is '{json_data.get("name")}' """
                f"it should be '{self.name}'"
            )
        if self.type != json_data.get("type"):
            raise KhiopsJSONError(
                f"""'name' field is '{json_data.get("type")}' """
                f"it should be '{self.type}'"
            )

        # Initialize clusters
        if "clusters" not in json_data:
            raise KhiopsJSONError("'clusters' key not found")
        for json_cluster in json_data["clusters"]:
            cluster = CoclusteringCluster(json_cluster)
            self.clusters.append(cluster)
            self._clusters_by_name[cluster.name] = cluster

        # Link the clusters according to their hierarchy
        # and link the leaf cluster to their parts
        for cluster in self.clusters:
            # Link leaf cluster to its part
            if cluster.is_leaf:
                cluster.leaf_part = self.get_part(cluster.name)

            # Root cluster case: Set the dimension's root cluster reference
            if cluster.parent_cluster_name == "":
                self.root_cluster = cluster
            # Non root case: Link the parent to the current cluster
            else:
                # Lookup the parent
                cluster.parent_cluster = self.get_cluster(cluster.parent_cluster_name)

                # If parent's first child empty: Set current cluster to it
                if cluster.parent_cluster.child_cluster1 is None:
                    cluster.parent_cluster.child_cluster1 = cluster
                # If not: Set current cluster to the second child
                else:
                    cluster.parent_cluster.child_cluster2 = cluster
        return self

    def get_part(self, part_name):
        """Returns a part of the dimension given the part's name

        Parameters
        ----------
        part_name : str
            Name of the part.

        Returns
        -------
        `CoclusteringDimensionPart`
            The part with the specified name.

        Raises
        ------
        `KeyError`
            If there is no part with the specified name.
        """
        return self._parts_by_name[part_name]

    def get_cluster(self, cluster_name):
        """Returns the specified cluster

        Parameters
        ----------
        cluster_name : str
            Name of the cluster.

        Returns
        -------
        `CoclusteringCluster`
            The specified cluster.

        Raises
        ------
        `KeyError`
            If there is no cluster with the specified name.
        """
        return self._clusters_by_name[cluster_name]

    def write_dimension_header_line(self, writer):
        """Writes the "dimensions" section header to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.write("Name\t")
        writer.write("Type\t")
        writer.write("Parts\t")
        writer.write("Initial parts\t")
        writer.write("Values\t")
        writer.write("Interest\t")
        writer.writeln("Description")

    def write_dimension_line(self, writer):
        """Writes the "dimensions" section line to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.write(f"{self.name}\t")
        writer.write(f"{self.type}\t")
        writer.write(f"{self.part_number}\t")
        writer.write(f"{self.initial_part_number}\t")
        writer.write(f"{self.value_number}\t")
        writer.write(f"{self.interest}\t")
        writer.writeln(self.description)

    def write_hierarchy(self, writer):
        """Writes the "hierarchy" section to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Write header
        writer.writeln("")
        writer.writeln(f"Hierarchy\t{self.name}")

        # Write the hierarchy for each cluster
        for i, cluster in enumerate(self.clusters):
            if i == 0:
                cluster.write_hierarchy_header_line(writer)
            cluster.write_hierarchy_line(writer)

    def write_composition(self, writer):
        """Writes the "composition" section to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Write only categorical dimensions
        if self.type == "Categorical":
            # Write header
            writer.writeln("")
            writer.writeln(f"Composition\t{self.name}")
            writer.write("Cluster\t")
            writer.write("Value\t")
            writer.write("Frequency\t")
            writer.writeln("Typicality")

            # Write value groups
            for value_group in self.parts:
                # Write value's information
                for value in value_group.values:
                    writer.write(f"{value_group.cluster_name}\t")
                    writer.write(f"{value.value}\t")
                    writer.write(f"{value.frequency}\t")
                    writer.writeln(str(value.typicality))

                # Write special value for the default part
                if value_group.is_default_part:
                    writer.write(f"{value_group.cluster_name}\t")
                    writer.write(" * \t")
                    writer.write(f"{0}\t")
                    writer.writeln(str(0))

    def write_annotation(self, writer):
        """Writes the "annotation" section to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Write header
        writer.writeln("")
        writer.writeln(f"Annotation\t{self.name}")

        # Write annotations for each cluster
        for i, cluster in enumerate(self.clusters):
            if i == 0:
                cluster.write_annotation_header_line(writer)
            cluster.write_annotation_line(writer)

    def needs_annotation_report(self):
        """Status about the annotation report

        Returns
        -------
        bool
            True if the "annotation" section is reported
        """
        # Check if any cluster has a non empty description
        for cluster in self.clusters:
            if cluster.short_description != "" or cluster.description != "":
                return True
        return False

    def write_hierarchy_structure_report_file(self, report_file_path):
        """Writes the hierarchical structure of the clusters to a file

        This method is mainly a test of the encoding of the cluster hierarchy.

        Parameters
        ----------
        report_file_path : str
            Path of the output file.
        """
        with open(report_file_path, "wb") as report_file:
            writer = KhiopsOutputWriter(report_file)
            writer.writeln(f"Hierarchical structure\t{self.name}")
            self.root_cluster.write_hierarchy_structure_report(writer)


class CoclusteringDimensionPart:
    """An element of a partition of a dimension

    Abstract class.

    Parameters
    ----------
    json_data: dict, optional
        See child classes for specific information about this parameter.

    Attributes
    ----------
    cluster_name : str
        Name of the cluster to which this part belongs.
    """

    def __init__(self, json_data=None):
        """See class docstring"""
        # Check the type of json_data
        if json_data is not None and not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Transform to an empty dictionary if json_data is not specified
        if json_data is None:
            json_data = {}
        # Otherwise check the validity of json_data
        elif "cluster" not in json_data:
            raise KhiopsJSONError("'cluster' key not found")

        # Cluster name of the part (only referenced by a leaf cluster of a hierarchy)
        self.cluster_name = json_data.get("cluster", "")


class CoclusteringDimensionPartInterval(CoclusteringDimensionPart):
    """An interval of a numerical partition

    Parameters
    ----------
    json_data : dict, optional
        Python dictionary representing an element of type "Numerical" of the list at the
        ``dimensionPartitions`` field of a Khiops Coclustering JSON report file. If not
        specified it returns an empty instance.

    Raises
    ------
    `.KhiopsJSONError`
        If ``json_data`` does not contain a "cluster" key.


    Attributes
    ----------
    cluster_name : str
        Name of the cluster containing this interval.
    lower_bound : float
        Lower bound of the interval.
    upper_bound : float
        Upper bound of the interval.
    is_missing : bool
        True if the instance's represent the missing values. In this case
        ``lower_bound`` and ``upper_bound`` are set to ``None``.
    is_left_open : bool
        True if the interval is unbounded below ``lower_bound`` may contain the minimum
        value of the training data.
    is_right_open : bool
        True if the interval is unbounded above ``upper_bound`` may contain the maximum
        value of training data.
    """

    def __init__(self, json_data=None):
        """See class docstring"""
        # Initialize base class
        super().__init__(json_data=json_data)

        # Transform to an empty dictionary if json_data is not specified
        if json_data is None:
            json_data = {}
        # Otherwise check that the 'bounds' key exists
        else:
            if "bounds" not in json_data:
                raise KhiopsJSONError("'bounds' key not found")
            elif not isinstance(json_data["bounds"], list):
                raise KhiopsJSONError("'bounds' key must be a list")
            elif len(json_data["bounds"]) != 2:
                raise KhiopsJSONError("'bounds' key must be a list of length 2")

        # Initialize a "true" interval
        json_bounds = json_data.get("bounds")
        if json_bounds:
            self.is_missing = False
            self.lower_bound = json_bounds[0]
            self.upper_bound = json_bounds[1]
        # Initialize a "missing part" interval
        else:
            self.is_missing = True
            self.lower_bound = None
            self.upper_bound = None

        # Initialize the 'open' flags (initialized by another class)
        self.is_left_open = False
        self.is_right_open = False

    def __str__(self):
        """Returns a human-readable string representation"""
        if self.is_missing:
            label = "Missing"
        else:
            if self.is_left_open:
                label = "]-inf"
            else:
                label = f"]{self.lower_bound}"
            if self.is_right_open:
                label += ";+inf["
            else:
                label += f";{self.upper_bound}]"
        return label

    def part_type(self):
        """Part type of this instance

        Returns
        -------
        str
            Only possible value: "Interval".
        """
        return "Interval"


class CoclusteringDimensionPartValueGroup(CoclusteringDimensionPart):
    """A value group of a categorical partition

    Parameters
    ----------
    json_data : dict, optional
        Python dictionary representing an element of type "Categorical" of the list at
        the ``dimensionPartitions`` field of a Khiops Coclustering JSON report file. If
        None it returns an empty instance.

    Raises
    ------
    `.KhiopsJSONError`
        If ``json_data`` does not contain a "cluster" key.

    Attributes
    ----------
    cluster_name : str
        Name of the cluster containing this group.
    values : list of `CoclusteringDimensionPartValue`
        The singleton parts composing this group part.
    is_default_part : bool
        True if the instance represents the "unknown values" group.
    """

    def __init__(self, json_data=None):
        """Constructs an instance from a python JSON object"""
        # Initialize base class
        super().__init__(json_data=json_data)

        # Transform to an empty dictionary if json_data is not specified
        if json_data is None:
            json_data = {}
        # Otherwise raise an error if the relevant keys are not found
        else:
            mandatory_keys = ("values", "valueFrequencies", "valueTypicalities")
            for key in mandatory_keys:
                if key not in json_data:
                    raise KhiopsJSONError(f"'{key}' key not found")
                elif not isinstance(json_data[key], list):
                    raise KhiopsJSONError(f"'{key}' key must be a list")
                elif len(json_data[key]) != len(json_data["values"]):
                    raise KhiopsJSONError(
                        f"'{key}' key list must have the same length as 'values'"
                    )

        # Initialize value list
        self.values = []
        json_values = json_data.get("values", [])
        json_value_frequencies = json_data.get("valueFrequencies", [])
        json_value_typicalities = json_data.get("valueTypicalities", [])
        for i, json_value in enumerate(json_values):
            value = CoclusteringDimensionPartValue()
            self.values.append(value)
            value.value = json_value
            value.frequency = json_value_frequencies[i]
            value.typicality = json_value_typicalities[i]

        # Initialize default values (set for real from another class)
        self.is_default_part = False

    def __str__(self):
        """Returns a human-readable string representation"""
        label = "{"
        for i, value in enumerate(self.values):
            if i > 0:
                label += ", "
            if i == 3:
                label += "..."
                break
            label += value.value
        label += "}"
        return label

    def part_type(self):
        """Part type of this instance

        Returns
        -------
        str
            Only possible value: "Value group".
        """
        return "Value group"


class CoclusteringDimensionPartValue:
    """A specific value of a variable in a dimension value group.

    .. note::
        This class has only a no-parameter constructor initializing an instance with the
        default values.

    Attributes
    ----------
    value : str
        String representation of the value.
    frequency : int
        Number of individuals having this value.
    typicality : float
        Indicates how much the value is representative of the cluster. Ranges from 0 to
        1, 1 being completely representative.
    """

    def __init__(self):
        """See class doctstring"""
        self.value = ""
        self.frequency = 0
        self.typicality = 0


class CoclusteringCluster:
    """A cluster in a coclustering dimension hierarchy

    Parameters
    ----------
    json_data : dict, optional
        JSON data of an element of the list at the ``dimensionHierarchies`` field within
        the ``coclusteringReport`` field of a Khiops Coclustering JSON report file. If
        not specified it returns an empty instance.

    Attributes
    ----------
    name : str
        Name of the cluster.
    parent_cluster_name : str
        Name of the parent cluster.
    frequency : int
        Number of individuals in the cluster.
    interest : float
        The cluster's interest/informativeness.
    hierarchical_level : float
        A measure interpretable as the distance of the cluster to the root. Between 0
        and 1.
    rank : int
        Rank of clusters in the top-down list of clusters, with the smallest ranks at
        the top.
    hierarchical_rank : int
        Rank of clusters in the hierarchy, with the smallest ranks being the closest
        from the root of the hierarchy.
    is_leaf : bool
        ``True`` if the cluster is a leaf of the hierarchy.
    short_description : str
        Succinct cluster description.
    description : str
        Cluster description.
    leaf_part : `CoclusteringDimensionPart`
        On a leaf cluster: Its unique associated partition element. Otherwise ``None``.
    parent_cluster : `CoclusteringCluster`
        On a non-root cluster: Its unique parent cluster. Otherwise ``None``.
    child_cluster1 : `CoclusteringCluster`
        On a non-leaf cluster : The first child cluster. Otherwise ``None``.
    child_cluster2 : `CoclusteringCluster`
        On a non-leaf cluster : The second child cluster. Otherwise ``None``.
    """

    def __init__(self, json_data=None):
        """See class docstring"""
        # Check the type of json_data
        if json_data is not None and not isinstance(json_data, dict):
            raise TypeError(type_error_message("json_data", json_data, dict))

        # Transform to an empty dictionary if json_data is not specified
        if json_data is None:
            json_data = {}
        # Otherwise check the "cluster" and "parentCluster" keys
        else:
            if "cluster" not in json_data:
                raise KhiopsJSONError("'cluster' key not found")
            elif "parentCluster" not in json_data:
                raise KhiopsJSONError("'parentCluster' key not found")

        # Initialize attributes
        self.name = json_data.get("cluster", "")
        self.parent_cluster_name = json_data.get("parentCluster", "")
        self.frequency = json_data.get("frequency", 0)
        self.interest = json_data.get("interest", 0.0)
        self.hierarchical_level = json_data.get("hierarchicalLevel", 0.0)
        self.rank = json_data.get("rank", 0)
        self.hierarchical_rank = json_data.get("hierarchicalRank", 0)
        self.is_leaf = json_data.get("isLeaf", False)
        self.short_description = json_data.get("shortDescription", "")
        self.description = json_data.get("description", "")

        # Link to child clusters, None for the leaves of the hierarchy
        # The user must specify the CoclusteringCluster references parent_cluster
        # and child_cluster that link this instance to the hierarchy
        self.parent_cluster = None
        self.child_cluster1 = None
        self.child_cluster2 = None
        self.leaf_part = None

    def write_hierarchy_header_line(self, writer):
        """Writes the "hierarchy" section's header to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.write("Cluster\t")
        writer.write("ParentCluster\t")
        writer.write("Frequency\t")
        writer.write("Interest\t")
        writer.write("HierarchicalLevel\t")
        writer.write("Rank\t")
        writer.writeln("HierarchicalRank")

    def write_hierarchy_line(self, writer):
        """Writes a line of the "hierarchy" section to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.write(f"{self.name}\t")
        writer.write(f"{self.parent_cluster_name}\t")
        writer.write(f"{self.frequency}\t")
        writer.write(f"{self.interest}\t")
        writer.write(f"{self.hierarchical_level}\t")
        writer.write(f"{self.rank}\t")
        writer.writeln(str(self.hierarchical_rank))

    def write_annotation_header_line(self, writer):
        """Writes the "annotation" section's header to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.write("Cluster\t")
        writer.write("Expand\t")
        writer.write("Selected\t")
        writer.write("ShortDescription\t")
        writer.writeln("Description")

    def write_annotation_line(self, writer):
        """Writes a line of the "annotation" section to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        writer.write(f"{self.name}\t")
        # TODO: Why "Expand" and "Selected" are not available?
        writer.write("FALSE\t")
        writer.write("FALSE\t")
        writer.write(f"{self.short_description}\t")
        writer.writeln(self.description)

    def write_hierarchy_structure_report(self, writer):
        """Writes the hierarchical structure from this instance to a writer object

        This method is mainly a test of the encoding of the cluster hierarchy.

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer for the report file.
        """
        # Write first child cluster
        if self.child_cluster1 is not None:
            self.child_cluster1.write_hierarchy_structure_report(writer)

        # Write current cluster
        writer.write(" " * self.hierarchical_rank)
        writer.write(self.name)
        writer.writeln("")

        # Write second child cluster
        if self.child_cluster2 is not None:
            self.child_cluster2.write_hierarchy_structure_report(writer)


class CoclusteringCell:
    """A coclustering cell

    .. note::
        This class has only a no-parameter constructor initializing an instance with the
        default values.

    Attributes
    ----------
    parts : list of `CoclusteringDimensionPart`
        Parts for each coclustering dimension.
    frequency : int
        Frequency of this cell.
    """

    def __init__(self):
        """Constructs an instance with default attribute values"""
        self.parts = []
        self.frequency = 0

    def write_line(self, writer):
        """Writes a line of the instance's report to a writer object

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Output writer.
        """
        for part in self.parts:
            writer.write(f"{part.cluster_name}\t")
        writer.writeln(str(self.frequency))



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""General helper functions"""

import itertools

from sklearn.model_selection import train_test_split

from khiops.core.internals.common import is_dict_like, type_error_message
from khiops.sklearn.dataset import Dataset, FileTable

# Note: We build the splits with lists and itertools.chain avoid pylint warning about
# unbalanced-tuple-unpacking. See issue https://github.com/pylint-dev/pylint/issues/5671


def train_test_split_dataset(ds_spec, y=None, test_size=0.25, **kwargs):
    """Splits a multi-table dataset spec into train and test

    Parameters
    ----------
    ds_spec : ``dict``
         A ``dict`` multi-table dataset specification (see :doc:`/multi_table_primer`).
         Only Pandas, NumPy, SciPy tables are accepted in the spec.
    y : :external:term:`array-like` of size (n_samples,) , optional
        The target values. ``n_samples`` is the number of rows of the main table in
        ``ds_spec``.
    test_size : float, default 0.25
        The proportion of the dataset (between 0.0 and 1.0) to be included in the test
        split.
    ... :
        Other optional parameters for `sklearn.model_selection.train_test_split`

    Raises
    ------
    `TypeError`
        If ``ds_spec`` is not dict-like.

    `ValueError`
        If the tables in ``ds_spec`` are file-paths.

    Examples
    --------
    See the following functions of the ``samples_sklearn.py`` documentation script:
        - `samples_sklearn.khiops_classifier_multitable_star`
        - `samples_sklearn.khiops_classifier_multitable_snowflake`

    """
    # Check the types
    if not is_dict_like(ds_spec):
        raise TypeError(type_error_message("ds_spec", ds_spec, "dict-like"))

    # Build the dataset for the feature table
    ds = Dataset(ds_spec)

    # Check the table type of the dataset
    if isinstance(ds.main_table, FileTable):
        raise ValueError(
            "Only Pandas, NumPy, SciPy sparse datasets may be used in this method."
        )

    # Obtain the keys for the other test_train_split function
    sklearn_split_params = {}
    for param in ("train_size", "random_state", "shuffle", "stratify"):
        if param in kwargs:
            sklearn_split_params[param] = kwargs[param]

    # Perform the split with and without target
    train_ds, test_ds, train_target_column, test_target_column = (
        _train_test_split_in_memory_dataset(
            ds,
            y,
            test_size=test_size,
            sklearn_split_params=sklearn_split_params,
        )
    )

    # Create the return tuple
    # Note:
    #     We use `itertools.chain` to avoid pylint false positive about
    #     unbalanced-tuple-unpacking. This warning appears when calling the function so
    #     users would be warned. To remove when the following issue is fixed:
    #     https://github.com/pylint-dev/pylint/issues/5671
    if y is None:
        split = itertools.chain((train_ds.to_spec(), test_ds.to_spec()))
    else:
        split = itertools.chain(
            (
                train_ds.to_spec(),
                test_ds.to_spec(),
                train_target_column,
                test_target_column,
            )
        )

    return split


def _train_test_split_in_memory_dataset(ds, y, test_size, sklearn_split_params=None):
    # Create shallow copies of the feature dataset
    train_ds = ds.copy()
    test_ds = ds.copy()

    # Split the main table and the target (if any)
    if y is None:
        train_ds.main_table.data_source, test_ds.main_table.data_source = (
            train_test_split(
                ds.main_table.data_source, test_size=test_size, **sklearn_split_params
            )
        )
        train_target_column = None
        test_target_column = None
    else:
        (
            train_ds.main_table.data_source,
            test_ds.main_table.data_source,
            train_target_column,
            test_target_column,
        ) = train_test_split(
            ds.main_table.data_source,
            y,
            test_size=test_size,
            **sklearn_split_params,
        )

    # Split the secondary tables tables
    # Note: The tables are traversed in BFS
    todo_relations = [
        relation for relation in ds.relations if relation[0] == ds.main_table.name
    ]
    while todo_relations:
        current_parent_table_name, current_child_table_name, _ = todo_relations.pop(0)
        for relation in ds.relations:
            parent_table_name, _, _ = relation
            if parent_table_name == current_child_table_name:
                todo_relations.append(relation)

        for new_ds in (train_ds, test_ds):
            origin_child_table = ds.get_table(current_child_table_name)
            new_child_table = new_ds.get_table(current_child_table_name)
            new_parent_table = new_ds.get_table(current_parent_table_name)
            new_parent_key_cols_df = new_parent_table.data_source[new_parent_table.key]
            new_child_table.data_source = new_parent_key_cols_df.merge(
                origin_child_table.data_source, on=new_parent_table.key
            )

    return train_ds, test_ds, train_target_column, test_target_column



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Pure Python library for the Khiops AutoML suite"""
from khiops.core.analysis_results import *
from khiops.core.api import *
from khiops.core.coclustering_results import *
from khiops.core.dictionary import *
from khiops.core.exceptions import *
from khiops.core.helpers import *
from khiops.core.internals.runner import get_runner, set_runner



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""check_database task family"""
from khiops.core.internals import task as tm
from khiops.core.internals.types import (
    BoolType,
    DictType,
    FloatType,
    IntType,
    StringLikeType,
)

# Disable long lines to have readable scenarios
# pylint: disable=line-too-long
TASKS = [
    tm.KhiopsTask(
        "check_database",
        "khiops",
        "10.0.0",
        [
            ("dictionary_file_path", StringLikeType),
            ("dictionary_name", StringLikeType),
            ("data_table_path", StringLikeType),
        ],
        [
            ("detect_format", BoolType, True),
            ("header_line", BoolType, True),
            ("field_separator", StringLikeType, ""),
            ("sample_percentage", FloatType, 100.0),
            ("sampling_mode", StringLikeType, "Include sample"),
            ("selection_variable", StringLikeType, ""),
            ("selection_value", StringLikeType, ""),
            ("additional_data_tables", DictType(StringLikeType, StringLikeType), None),
            ("max_messages", IntType, 20),
        ],
        ["dictionary_file_path", "data_table_path", "additional_data_tables"],
        # fmt: off
        """
        // Dictionary file and class settings
        ClassManagement.OpenFile
        ClassFileName __dictionary_file_path__
        OK
        ClassManagement.ClassName __dictionary_name__

        // Train database settings
        TrainDatabase.DatabaseFiles.List.Key __dictionary_name__
        TrainDatabase.DatabaseFiles.DataTableName __data_table_path__
        __DICT__
        __additional_data_tables__
        TrainDatabase.DatabaseFiles.List.Key
        TrainDatabase.DatabaseFiles.DataTableName
        __END_DICT__
        TrainDatabase.HeaderLineUsed __header_line__
        TrainDatabase.FieldSeparator __field_separator__
        __OPT__
        __detect_format__
        TrainDatabase.DatabaseFormatDetector.DetectFileFormat
        __END_OPT__
        TrainDatabase.SampleNumberPercentage __sample_percentage__
        TrainDatabase.SamplingMode __sampling_mode__
        TrainDatabase.SelectionAttribute __selection_variable__
        TrainDatabase.SelectionValue __selection_value__

        // Log messages limit
        AnalysisSpec.SystemParameters.MaxErrorMessageNumberInLog __max_messages__

        // Execute check database
        LearningTools.CheckData
        """,
        # fmt: on
    ),
]



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Miscellaneous utility tools

.. warning::
    The entry point functions in this module use `sys.exit`. They are not designed to be
    called from another program or python shell.
"""
import argparse
import os
import pathlib
import shutil
import sys
import tempfile
import urllib.request
import warnings
import zipfile

import khiops.core as kh
from khiops.core.internals.common import deprecation_message
from khiops.samples import samples as samples_core

# We deactivate the warnings to not show a deprecation warning from sklearn
# pylint: disable=wrong-import-position
warnings.simplefilter("ignore")
from khiops.samples import samples_sklearn

warnings.resetwarnings()
# pylint: enable=wrong-import-position

# Note: We dont include these tools in coverage


def pk_status_entry_point():  # pragma: no cover
    """Entry point of the pk-status command

    **Deprecated** will be removed in Khiops 11, use `kh_status_entry_point` instead.
    """
    warnings.warn(
        deprecation_message(
            deprecated_feature="pk-status",
            replacement="kh-status",
            deadline_version="11.0.0",
        )
    )
    kh_status_entry_point()


def kh_status_entry_point():  # pragma: no cover
    """Entry point of the kh-status command"""
    status_code = kh.get_runner().print_status()
    sys.exit(status_code)


def kh_samples_entry_point():  # pragma: no cover
    """Entry point of the kh-samples command"""
    try:
        # For technical reasons, the arguments to the script can only be read
        # from `sys.argv`, as entry point script functions cannot have parameters
        args = sys.argv[1:]

        # Check which samples to execute
        if len(args) > 0 and args[0] in ("core", "sklearn"):
            submodule = args.pop(0)
        # By default, execute khiops.core samples
        else:
            submodule = "core"

        # Execute the required samples
        if submodule == "core":
            samples_module = samples_core
        else:
            samples_module = samples_sklearn

        argument_parser = samples_module.build_argument_parser(
            prog="kh-samples [core|sklearn]",
            description=(
                "Executes the sample code snippets of the Khiops Python library"
            ),
        )
        samples_module.execute_samples(argument_parser.parse_args(args))
    except kh.KhiopsRuntimeError as error:
        print(f"khiops engine error: {error}")
        sys.exit(1)


# Samples version: To be updated when khiops-samples does
DEFAULT_SAMPLES_VERSION = "10.2.4"


def kh_download_datasets_entry_point():
    """Entry point for the download samples helper script"""
    samples_dir = pathlib.Path.home() / "khiops_data" / "samples"
    arg_parser = argparse.ArgumentParser(
        prog="kh-download-datasets",
        description=f"Downloads the Khiops samples dataset to {samples_dir}.",
    )
    arg_parser.add_argument(
        "-v",
        "--version",
        default=DEFAULT_SAMPLES_VERSION,
        help="Sample datasets version",
    )
    arg_parser.add_argument(
        "-f",
        "--force-overwrite",
        default=False,
        action="store_true",
        help=f"Overwrites any existent directory at {samples_dir}",
    )
    args = arg_parser.parse_args()
    with warnings.catch_warnings(record=True) as caught_warnings:
        download_datasets(
            force_overwrite=args.force_overwrite,
            version=args.version,
            _called_from_shell=True,
        )
        if caught_warnings:
            for warning in caught_warnings:
                print(f"warning: {warning.message}")
    sys.exit(0)


def download_datasets(
    force_overwrite=False, version=DEFAULT_SAMPLES_VERSION, _called_from_shell=False
):
    """Downloads the Khiops sample datasets for a given version

    The datasets are downloaded to:
        - Windows: ``%USERPROFILE%\\khiops_data\\samples``
        - Linux/macOS: ``$HOME/khiops_data/samples``

    Parameters
    ==========
    force_overwrite : bool, default ``False``
        If ``True`` it always overwrites the local samples directory even if it exists.
    version : str, default "10.2.0"
        The version of the samples datasets.
    """
    # Note: The hidden parameter _called_from_shell is just to change the user messages.

    # Check if the home sample dataset location is available and build it if necessary
    samples_dir = pathlib.Path.home() / "khiops_data" / "samples"
    if samples_dir.exists() and not force_overwrite:
        if _called_from_shell:
            instructions = "Execute with '--force-overwrite' to overwrite it"
        else:
            instructions = "Set 'force_overwrite=True' to overwrite it"
        warnings.warn(
            "Download not executed since the sample datasets directory "
            f"already exists. {instructions}. Path: {samples_dir}"
        )
    else:
        # Create the samples dataset directory
        if samples_dir.exists():
            shutil.rmtree(samples_dir)
        os.makedirs(samples_dir, exist_ok=True)

        # Set the sample dataset zip URL
        samples_repo_url = "https://github.com/KhiopsML/khiops-samples"
        samples_zip_file = f"khiops-samples-{version}.zip"
        samples_zip_url = (
            f"{samples_repo_url}/releases/download/{version}/{samples_zip_file}"
        )

        # Download the sample zip file and extracted to the home dataset dir
        print(f"Downloading samples from {samples_zip_url}")
        with tempfile.NamedTemporaryFile() as temp_zip_file, urllib.request.urlopen(
            samples_zip_url
        ) as zip_request:
            temp_zip_file.write(zip_request.read())
            temp_zip_file.seek(0)
            with zipfile.ZipFile(temp_zip_file) as temp_zip:
                temp_zip.extractall(samples_dir)
            print(f"Samples dataset successfully downloaded to {samples_dir}")



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Core API functions abstractions"""
import textwrap

from khiops.core.internals.common import (
    create_unambiguous_khiops_path,
    type_error_message,
)
from khiops.core.internals.io import encode_file_path
from khiops.core.internals.scenario import ConfigurableKhiopsScenario
from khiops.core.internals.types import (
    AbstractDictType,
    AbstractListType,
    DictType,
    KhiopsTaskArgumentType,
    StringLikeType,
)
from khiops.core.internals.version import KhiopsVersion


def encode_path_valued_arg(arg, arg_name, arg_type):
    """Encodes an argument containing paths (str or dict)

    Parameters
    ----------
    arg : str or dict
        The value of a path valued argument.
    arg_name : str
        The name of the argument.
    arg_type : `.KhiopsTaskArgumentType`
        The type of the argument. Must be either `.StringLikeType` or
        `.DictType`  `.StringLikeType` , `.StringLikeType` .

    See Also
    --------
    .encode_path_valued_arg : The underlying transformation function.

    """
    if not issubclass(arg_type, StringLikeType) and not issubclass(
        arg_type, DictType(StringLikeType, StringLikeType)
    ):
        raise TypeError(type_error_message(arg_name, arg, str, bytes, "dict-str-str"))
    if issubclass(arg_type, StringLikeType):
        encoded_arg = encode_file_path(arg)
    else:
        encoded_arg = {}
        for key in arg.keys():
            encoded_arg[key] = encode_file_path(arg[key])
    return encoded_arg


class KhiopsTask:
    """Specification of a Khiops Task

    Parameters
    ----------
    name : str
        Name of the task.
    tool_name : str
        Name of the tool to execute the task (ex. "khiops", "khiops_coclustering").
    intro_version : `.KhiopsVersion` or str
        Khiops version where this task object was introduced.
    args_signature : list
        A list of 2-tuples for each mandatory parameter of the task, containing:
            - its name (str)
            - its type (`.KhiopsTaskArgumentType`)
    kwargs_signature : list
        A list of 3-tuples for each optional parameter of the task, containing:
            - its name (str)
            - its type (`.KhiopsTaskArgumentType`)
            - its default value
    path_valued_arg_names : list of str
        A list of the parameters that contain paths. They must be contained either in
        args_signature or kwargs_signature. The only accepted types are
        `.StringLikeType` or containers thereof.
    scenario_template : str
        A Khiops scenario template. The template mini-language is described above.

    Attributes
    ----------
    scenario : `.ConfigurableKhiopsScenario`
        Scenario object built from ``scenario_template``.
    """

    def __init__(
        self,
        name,
        tool_name,
        intro_version,
        args_signature,
        kwargs_signature,
        path_valued_arg_names,
        scenario_template,
    ):
        # Pre-initialization checks
        # Check the basic input types
        if not isinstance(name, str):
            raise TypeError(type_error_message("name", name, str))
        if not isinstance(tool_name, str):
            raise TypeError(type_error_message("tool_name", tool_name, str))
        if not isinstance(intro_version, (str, KhiopsVersion)):
            raise TypeError(type_error_message("version", intro_version, str))
        if not isinstance(args_signature, list):
            raise TypeError(type_error_message("args_signature", args_signature, list))
        if not isinstance(kwargs_signature, list):
            raise TypeError(
                type_error_message("kwargs_signature", kwargs_signature, list)
            )
        if not isinstance(path_valued_arg_names, list):
            raise TypeError(
                type_error_message("path_valued_arg_names", path_valued_arg_names, list)
            )
        if not isinstance(scenario_template, str):
            raise TypeError(
                type_error_message("scenario_template", scenario_template, str)
            )

        # Check the types of the contents of args
        for arg_index, arg_tuple in enumerate(args_signature):
            if not isinstance(arg_tuple, tuple):
                raise TypeError(
                    type_error_message(f"arg[{arg_index}] tuple", arg_tuple, tuple)
                )
            if len(arg_tuple) != 2:
                raise ValueError(f"arg[{arg_index}] tuple must have length 2")
            arg_name, arg_type = arg_tuple
            if not isinstance(arg_name, str):
                raise TypeError(type_error_message(f"arg[{arg_index}])", arg_name, str))
            if not issubclass(arg_type, KhiopsTaskArgumentType):
                raise TypeError(
                    type_error_message(
                        f"arg[{arg_index}]'s type", arg_type, KhiopsTaskArgumentType
                    )
                )

        # Check the types of the contents of kwargs
        for kwarg_index, kwarg_tuple in enumerate(kwargs_signature):
            if not isinstance(kwarg_tuple, tuple):
                raise TypeError(
                    type_error_message(
                        f"kwarg[{kwarg_index}] tuple", kwarg_tuple, tuple
                    )
                )
            if len(kwarg_tuple) != 3:
                raise ValueError(f"kwarg[{kwarg_index}] tuple must have length 3")
            kwarg_name, kwarg_type, kwarg_default = kwarg_tuple
            if not isinstance(kwarg_name, str):
                raise TypeError(
                    type_error_message(f"kwarg[{kwarg_index}])", kwarg_name, str)
                )
            if not issubclass(kwarg_type, KhiopsTaskArgumentType):
                raise TypeError(
                    type_error_message(
                        f"kwarg[{kwarg_index}]) type",
                        kwarg_type,
                        KhiopsTaskArgumentType,
                    )
                )
            if kwarg_default is not None:
                if not kwarg_type.is_of_this_type(kwarg_default):
                    raise TypeError(
                        type_error_message(
                            f"kwarg[{kwarg_index}]) default",
                            kwarg_default,
                            kwarg_type.short_name(),
                            "None",
                        )
                    )
                # We do not admit bytes default values for KhString
                if issubclass(kwarg_type, StringLikeType) and isinstance(
                    kwarg_default, bytes
                ):
                    raise TypeError(
                        f"kwarg[{kwarg_index}]) string-like default value "
                        "cannot be of type 'bytes'"
                    )

        # Check the types of path_valued_arg_names
        for arg_index, arg_name in enumerate(path_valued_arg_names):
            if not isinstance(arg_name, str):
                raise TypeError(
                    type_error_message(
                        f"path_valued_arg_names[{arg_index}]", arg_name, str
                    )
                )

        # Initialize the public members
        self.name = name
        self.tool_name = tool_name
        if isinstance(intro_version, str):
            self.intro_version = KhiopsVersion(intro_version)
        else:
            self.intro_version = intro_version
        self.args_signature = args_signature
        self.kwargs_signature = kwargs_signature
        self.scenario = ConfigurableKhiopsScenario(textwrap.dedent(scenario_template))
        self.path_valued_arg_names = path_valued_arg_names

        # Index the arguments by name
        self._args_signature_by_name = {
            arg_signature[0]: arg_signature for arg_signature in self.args_signature
        }
        self._kwargs_signature_by_name = {
            kwarg_signature[0]: kwarg_signature
            for kwarg_signature in self.kwargs_signature
        }

        # Post-initialization checks
        # Check that the path_valued_arg_names are contained in either args or kwargs
        all_arg_names = list(self._args_signature_by_name.keys()) + list(
            self._kwargs_signature_by_name.keys()
        )
        for path_valued_arg_name in path_valued_arg_names:
            if path_valued_arg_name not in all_arg_names:
                raise ValueError(
                    f"Path argument '{path_valued_arg_name}' not found "
                    "in args nor kwargs"
                )

        # Check the keyword coherence
        all_scenario_arg_names = [f"__{arg_name}__" for arg_name in all_arg_names]
        self.scenario.check_keyword_completeness(all_scenario_arg_names)

        # Check the coherence of path_valued_arg_names
        for arg_name in self.path_valued_arg_names:
            if (
                arg_name not in self._args_signature_by_name
                and not self._kwargs_signature_by_name
            ):
                raise ValueError(f"Path argument '{arg_name}' not found in signature")
            if arg_name in self._args_signature_by_name:
                _, arg_type = self._args_signature_by_name[arg_name]
            else:
                _, arg_type, _ = self._kwargs_signature_by_name[arg_name]

            if not issubclass(arg_type, StringLikeType) and not issubclass(
                arg_type, DictType(StringLikeType, StringLikeType)
            ):
                raise TypeError(
                    f"Path argument '{arg_name}' should be string-like or "
                    "dict of string-like pairs"
                )

    def write_execution_scenario(self, writer, args):
        """Writes a task's execution scenario file contents for the specified arguments

        Parameters
        ----------
        writer : `.KhiopsOutputWriter`
            Writer object for the scenario file.
        args : dict
            Task arguments. Includes mandatory and optional arguments.
        """
        # Check that all the arguments are known
        for arg_name, arg in args.items():
            if (
                not arg_name in self._args_signature_by_name
                and not arg_name in self._kwargs_signature_by_name
            ):
                raise ValueError(f"Unknown argument '{arg_name}'")

        # Check that all mandatory arguments are present
        for arg_name in self._args_signature_by_name.keys():
            if not arg_name in args:
                raise ValueError(f"Missing mandatory argument '{arg_name}'")

        # Check the types of the arguments
        for arg_name, arg in args.items():
            if arg_name in self._args_signature_by_name:
                _, arg_type = self._args_signature_by_name[arg_name]
            else:
                assert arg_name in self._kwargs_signature_by_name
                _, arg_type, _ = self._kwargs_signature_by_name[arg_name]
            if (
                not issubclass(arg_type, (AbstractDictType, AbstractListType))
                or arg is not None
            ):
                arg_type.check(arg, arg_name)

        # Save the non-specified optional arguments
        absent_kwarg_names = []
        for arg_name in self._kwargs_signature_by_name:
            if arg_name not in args:
                absent_kwarg_names.append(arg_name)

        # Disambiguate path parameters affected of Khiops silently
        # changing its working directory to that of the data tables
        for path_arg_name in [
            "results_dir",
            "evaluation_report_path",
            "output_data_table_path",
        ]:
            if path_arg_name in args:
                args[path_arg_name] = create_unambiguous_khiops_path(
                    args[path_arg_name]
                )
        if (
            "output_additional_data_tables" in args
            and args["output_additional_data_tables"] is not None
        ):
            for data_path in args["output_additional_data_tables"].keys():
                args["output_additional_data_tables"][data_path] = (
                    create_unambiguous_khiops_path(
                        args["output_additional_data_tables"][data_path]
                    )
                )

        # Transform to string-like parameters
        # Path-valued parameters are encoded differently depending on the platform
        scenario_args = {}
        for arg_name, arg in args.items():
            if arg_name in self._args_signature_by_name:
                _, arg_type = self._args_signature_by_name[arg_name]
                scenario_arg = arg_type.to_scenario_arg(arg)
            else:
                assert arg_name in self._kwargs_signature_by_name
                _, arg_type, arg_default = self._kwargs_signature_by_name[arg_name]
                if arg is not None:
                    scenario_arg = arg_type.to_scenario_arg(arg)
                else:
                    scenario_arg = arg
            if arg_name in self.path_valued_arg_names and arg is not None:
                scenario_arg = encode_path_valued_arg(scenario_arg, arg_name, arg_type)
            scenario_args[f"__{arg_name}__"] = scenario_arg

        # Transform absent parameters (with their default value)
        for arg_name in absent_kwarg_names:
            _, arg_type, arg_default = self._kwargs_signature_by_name[arg_name]
            if arg_default is not None:
                scenario_arg = arg_type.to_scenario_arg(arg_default)
            else:
                scenario_arg = None
            scenario_args[f"__{arg_name}__"] = scenario_arg

        # Write the scenario file
        writer.writeln(f"// Scenario for task {self.name}")
        self.scenario.write(writer, scenario_args)
        writer.writeln("")
        writer.writeln(f"// End of scenario for task {self.name}")


class KhiopsTaskFamily:
    """Stores the versions of a task

    Parameters
    ----------
    task_name : str
        The name of the task defining this family.
    end_version : str or `.KhiopsVersion`
        The Khiops version where the support for this task ended.
    """

    def __init__(self, task_name, end_version=None):
        """See class docstring"""
        self.task_name = task_name
        self.end_version = end_version
        self.tasks = {}

    @property
    def end_version(self):
        """`.KhiopsVersion` : Khiops version where support for this task ended

        It is ``None`` if the task is still supported. May be set with either str or
        `.KhiopsVersion` or ``None``.
        """
        return self._end_version

    @end_version.setter
    def end_version(self, version):
        if not isinstance(version, (str, KhiopsVersion, type(None))):
            raise TypeError(
                type_error_message("version", version, str, KhiopsVersion, type(None))
            )
        if isinstance(version, str):
            self._end_version = KhiopsVersion(version)
        else:
            assert isinstance(version, (KhiopsVersion, type(None)))
            self._end_version = version

    @property
    def start_version(self):
        """`.KhiopsVersion` : Khiops version where the support for this task started

        Raises
        ------
        `ValueError`
            If there are no task registered in this family.
        """
        if not self.tasks:
            raise ValueError(
                "There are no registered versions " f"for task '{self.task_name}'"
            )
        start_version = self.all_intro_versions[-1]
        return start_version

    def register_task(self, task, overwrite=False):
        """Registers a task to this family

        Parameters
        ----------
        task : `.KhiopsTask`
            The task to be registered. Must have the same name of the family.
        overwrite : bool, default ``False``
            If ``True`` it does not raise an error if a task with the same version is
            already registered.

        Raises
        ------
        `ValueError`
            If:

            - The task name is not the same as that of this family.
            - The specified version already exists and overwrite is not set.
        """

        if task.name != self.task_name:
            raise ValueError(
                f"task name '{task.name}' different from "
                f"task family name '{self.task_name}'"
            )
        if task.intro_version in self.tasks and not overwrite:
            raise ValueError(
                f"Cannot replace task '{task.name}' version {task.intro_version} "
                "since it already exists (set overwrite to True to override)"
            )
        self.tasks[task.intro_version] = task

    def unregister_task(self, intro_version):
        """Unregister the task from the family with the specified introduction version

        Parameters
        ----------
        intro_version : str
            The introduction version of the task to be unregistered.

        Returns
        -------
        `KhiopsTask`
            The removed task object.

        Raises
        ------
        `ValueError`
            If there is no task with the specified introduction version.
        """
        if intro_version not in self.tasks:
            raise ValueError(
                f"Cannot unregister inexistent version {intro_version} "
                f"of task {self.task_name}"
            )
        return self.tasks.pop(intro_version)

    def get_task(self, target_version):
        """Returns latest task object compatible with the specified version

        Parameters
        ----------
        target_version : str or `.KhiopsVersion`
            The target Khiops version.

        Returns
        -------
        `.KhiopsTask`
            The latest task object compatible with the ``target_version``.

        Raises
        ------
        `TypeError`
            If target_version is not str or `.KhiopsVersion`.

        """
        # Check the argument type
        if not isinstance(target_version, (str, KhiopsVersion)):
            raise TypeError(type_error_message(target_version, str, KhiopsVersion))

        # Transform target_version to KhiopsVersion if necessary
        if isinstance(target_version, str):
            target_version = KhiopsVersion(target_version)

        # Fail if the target version is ahead the end_version
        if self.end_version is not None and target_version > self.end_version:
            raise ValueError(
                f"No compatible version for '{self.task_name}' "
                f"version {target_version}: "
                f"This task was available only up to version {self.end_version}"
            )

        # Search the newest compatible version of the function
        newest_compatible_version = None
        for intro_version in self.all_intro_versions:
            if target_version >= intro_version:
                newest_compatible_version = intro_version
                break

        # Fail if there is no compatible version
        if newest_compatible_version is None:
            raise ValueError(
                f"No compatible version for '{self.task_name}' "
                f"version {target_version}: "
                f"Earliest introduction version {self.all_intro_versions[-1]}"
            )

        return self.tasks[newest_compatible_version]

    @property
    def all_intro_versions(self):
        """list of `.KhiopsVersion` : A sorted list of the task introduction versions.

        The list is sorted in reverse order.
        """
        return sorted(list(self.tasks.keys()), reverse=True)

    @property
    def latest_intro_version(self):
        """`.KhiopsVersion` : The latest introduction version of the task.

        Raises
        ------
        ValueError
            If there are no task objects registered in this family.
        """
        if not self.tasks:
            raise ValueError(
                "There are no registered versions " f"for task '{self.task_name}'"
            )
        return self.all_intro_versions[0]


class KhiopsTaskRegistry:
    """Registry of Khiops tasks

    Tasks are indexed by name and version.
    """

    def __init__(self):
        self.task_families = {}

    def register_task(self, task):
        if not isinstance(task, KhiopsTask):
            raise TypeError(type_error_message("task", task, KhiopsTask))
        if task.name not in self.task_families:
            self.task_families[task.name] = KhiopsTaskFamily(task.name)
        self.task_families[task.name].register_task(task)

    def get_task(self, task_name, target_version):
        """Retrieves the latest task object for the specified task and version

        Parameters
        ----------
        task_name : str
            The name of the task to be retrieved.
        target_version : str or `.KhiopsVersion`
            The target Khiops version.

        Returns
        -------
        `.KhiopsTask`
            The latest task object for the specified task and version.

        Raises
        ------
        `ValueError`
            If there are no tasks registered with the specified name.
        """
        # Check the argument type
        if not isinstance(target_version, (str, KhiopsVersion)):
            raise TypeError(type_error_message(target_version, str, KhiopsVersion))

        # Check that the task family exists
        if task_name not in self.task_families:
            raise ValueError(f"There are no tasks named '{task_name}' registered")

        return self.task_families[task_name].get_task(target_version)

    def get_tasks(self, target_version):
        """Retrieves the latest Khiops tasks compatible with the specified version

        Parameters
        ----------
        target_version : str or `.KhiopsVersion`
            The target Khiops version for the tasks to be retrieved.

        Returns
        -------
        list of `.KhiopsTask`
            The list of compatible tasks with ``target_version``.
        """
        # Check the argument type
        if not isinstance(target_version, (str, KhiopsVersion)):
            raise TypeError(type_error_message(target_version, str, KhiopsVersion))

        # Transform the version to KhiopsVersion if necessary
        if isinstance(target_version, str):
            target_version = KhiopsVersion(target_version)

        # Fill the task list with the latests versions
        compatible_tasks = []
        for task_name in self.task_names:
            # Skip the task if the target version is ahead of its end version
            task_end_version = self.task_families[task_name].end_version
            if task_end_version is not None and task_end_version < target_version:
                continue

            # Add the task with the latest compatible version
            compatible_tasks.append(self.get_task(task_name, target_version))

        return compatible_tasks

    def get_task_end_version(self, task_name):
        """Returns the version where the support of the specified task ended

        Parameters
        ----------
        task_name : str
            Name of the task.

        Returns
        -------
        `.KhiopsVersion` or ``None``
            Either the end version for this task or ``None`` if it is still supported.

        Raises
        ------
        `TypeError`
            If ``task_name`` is not of type str.
        `ValueError`
            If there are no registered tasks with the specified name.
        """
        if not isinstance(task_name, str):
            raise TypeError(type_error_message("task_name", task_name, str))
        if task_name not in self.task_families:
            raise ValueError(f"Task family for '{task_name}' not registered")
        return self.task_families[task_name].end_version

    def set_task_end_version(self, task_name, end_version):
        """Sets the version where the support of the specified task ended

        Parameters
        ----------
        task_name : str
            Name of the task.
        end_version : str or `.KhiopsVersion`
            Version where the support of the specified task ended.

        Raises
        ------
        `TypeError`
            - If ``task_name`` is not of type str.
            - If ``end_version`` is not of type str or `.KhiopsVersion`.
        `ValueError`
            If there are no registered tasks with the specified name.
        """
        # Check input types
        # Note: `end_version` type check is deferred to `TaskFamily`
        if not isinstance(task_name, str):
            raise TypeError(type_error_message("task_name", task_name, str))
        if task_name not in self.task_families:
            raise ValueError(f"Task family for '{task_name}' not registered")
        self.task_families[task_name].end_version = end_version

    @property
    def latest_intro_version(self):
        """`.KhiopsVersion` : Latest introduction version overall tasks"""
        latest_intro_version = KhiopsVersion("1.0")
        for task_family in self.task_families:
            latest_intro_version = max(
                task_family.latest_intro_version, latest_intro_version
            )
        return latest_intro_version

    @property
    def task_names(self):
        """list of str: Names of all task families"""
        return list(self.task_families.keys())


# _REGISTRY is global but it is not assigned because it is already initialized
# pylint: disable=global-variable-not-assigned
_REGISTRY = KhiopsTaskRegistry()


def get_task_registry():
    """Direct access to the registry"""
    global _REGISTRY
    return _REGISTRY



######################################################################################
# Copyright (c) 2023-2025 Orange. All rights reserved.                               #
# This software is distributed under the BSD 3-Clause-clear License, the text of     #
# which is available at https://spdx.org/licenses/BSD-3-Clause-Clear.html or         #
# see the "LICENSE.md" file for more details.                                        #
######################################################################################
"""Classes for handling diverse data tables"""
import csv
import io
import json
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence

import numpy as np
import pandas as pd
import sklearn
from scipy import sparse as sp
from sklearn.utils import check_array
from sklearn.utils.validation import column_or_1d

import khiops.core as kh
import khiops.core.internals.filesystems as fs
from khiops.core.dictionary import VariableBlock
from khiops.core.exceptions import KhiopsRuntimeError
from khiops.core.internals.common import (
    deprecation_message,
    is_dict_like,
    is_list_like,
    type_error_message,
)

# Disable PEP8 variable names because of scikit-learn X,y conventions
# To capture invalid-names other than X,y run:
#   pylint --disable=all --enable=invalid-names dataset.py
# pylint: disable=invalid-name


def check_dataset_spec(ds_spec):
    """Checks that a dataset spec is valid

    Parameters
    ----------
    ds_spec : dict
        A specification of a multi-table dataset (see :doc:`/multi_table_primer`).

    Raises
    ------
    TypeError
        If there are objects of the spec with invalid type.
    ValueError
        If there are objects of the spec with invalid values.
    """
    # Check the spec type
    if not is_dict_like(ds_spec):
        raise TypeError(type_error_message("ds_spec", ds_spec, Mapping))

    # Check the "tables" field
    if "tables" not in ds_spec:
        raise ValueError("'tables' entry missing from dataset dict spec")
    if not is_dict_like(ds_spec["tables"]):
        raise TypeError(
            type_error_message("'tables' entry", ds_spec["tables"], Mapping)
        )
    if len(ds_spec["tables"]) == 0:
        raise ValueError("'tables' dictionary cannot be empty")
    for table_name, table_entry in ds_spec["tables"].items():
        _check_table_entry(table_name, table_entry)

    # Multi-table specific table checks
    if len(ds_spec["tables"]) > 1:
        _check_multitable_spec(ds_spec)

    # Check the 'format' field
    if "format" in ds_spec:
        _check_format_entry(ds_spec["format"])


def _check_table_entry(table_name, table_spec):
    if not isinstance(table_spec, tuple):
        raise TypeError(
            type_error_message(f"'{table_name}' table entry", table_spec, tuple)
        )
    if len(table_spec) != 2:
        raise ValueError(
            f"'{table_name}' table entry must have size 2, not {len(table_spec)}"
        )
    source, key = table_spec
    if not isinstance(source, (pd.DataFrame, sp.spmatrix, str)) and not hasattr(
        source, "__array__"
    ):
        raise TypeError(
            type_error_message(
                f"'{table_name}' table's source",
                source,
                "array-like",
                "scipy.sparse.spmatrix",
                str,
            )
        )
    _check_table_key(table_name, key)


def _check_table_key(table_name, key):
    if key is not None:
        if not is_list_like(key) and not isinstance(key, str):
            raise TypeError(
                type_error_message(f"'{table_name}' table's key", key, str, Sequence)
            )
        if len(key) == 0:
            raise ValueError(f"'{table_name}' table's key is empty")
        for column_name in key:
            if not isinstance(column_name, str):
                raise TypeError(
                    type_error_message(
                        f"'{table_name}' table's key column name",
                        column_name,
                        str,
                    )
                )


def _check_multitable_spec(ds_spec):
    # Check the main table
    if "main_table" not in ds_spec:
        raise ValueError(
            "'main_table' entry must be specified for multi-table datasets"
        )
    if not isinstance(ds_spec["main_table"], str):
        raise TypeError(
            type_error_message("'main_table' entry", ds_spec["main_table"], str)
        )
    if ds_spec["main_table"] not in ds_spec["tables"]:
        raise ValueError(
            f"A table entry with the main table's name ('{ds_spec['main_table']}') "
            f"must be present in the 'tables' dictionary"
        )

    # Check that all tables have non-None keys
    for table_name, (_, table_key) in ds_spec["tables"].items():
        if table_key is None:
            table_kind = "main" if ds_spec["main_table"] == table_name else "secondary"
            raise ValueError(
                f"key of {table_kind} table '{table_name}' is 'None': "
                "table keys must be specified in multi-table datasets"
            )

    # Check that all the tables have the same type as the main
    main_table_type = type(ds_spec["tables"][ds_spec["main_table"]][0])
    for table_name, (table_source, _) in ds_spec["tables"].items():
        if table_name != ds_spec["main_table"]:
            if not isinstance(table_source, main_table_type):
                raise ValueError(
                    f"Secondary table '{table_name}' has type "
                    f"'{type(table_source).__name__}' which is different from the "
                    f"main table's type '{main_table_type.__name__}'."
                )

    # If the 'relations' entry exists check it
    if "relations" in ds_spec:
        relations_spec = ds_spec["relations"]
    # Otherwise build a star schema relations spec and check it
    else:
        relations_spec = [
            (ds_spec["main_table"], table)
            for table in ds_spec["tables"].keys()
            if table != ds_spec["main_table"]
        ]
    _check_relations_entry(ds_spec["main_table"], ds_spec["tables"], relations_spec)


def _check_relations_entry(main_table_name, tables_spec, relations_spec):
    # Check the types and size of the relation entries
    if not is_list_like(relations_spec):
        raise TypeError(
            type_error_message("'relations' entry", relations_spec, "list-like")
        )
    for i, relation in enumerate(relations_spec, 1):
        # Check that the relation is a 2 or 3 tuple
        if not isinstance(relation, tuple):
            raise TypeError(type_error_message("Relation", relation, tuple))
        if len(relation) not in (2, 3):
            raise ValueError(f"A relation must be of size 2 or 3, not {len(relation)}")

        # Check the types of the tuple contents
        parent_table, child_table = relation[:2]
        if not isinstance(parent_table, str):
            raise TypeError(
                type_error_message(f"Relation #{i}'s parent table", parent_table, str)
            )
        if not isinstance(child_table, str):
            raise TypeError(
                type_error_message(f"Relation #{i}'s child table", child_table, str)
            )
        if len(relation) == 3 and not isinstance(relation[2], bool):
            raise TypeError(
                type_error_message(
                    f"Relation #{i} ({parent_table}, {child_table}) 1-1 flag",
                    relation[2],
                    bool,
                )
            )

    # Check structure and coherence with the rest of the spec
    parents_and_children = [relation[:2] for relation in relations_spec]
    for i, relation in enumerate(relations_spec, 1):
        parent_table, child_table = relation[:2]
        if parent_table == child_table:
            raise ValueError(
                f"Relation #{i}'s tables are equal: ({parent_table}, {child_table}). "
                "They must be different."
            )
        for table in (parent_table, child_table):
            if not table in tables_spec.keys():
                raise ValueError(
                    f"Relation #{i} ({parent_table}, {child_table}) contains "
                    f"non-existent table '{table}'. All relation tables must exist "
                    "in the 'tables' entry."
                )
        if parents_and_children.count(relation[:2]) > 1:
            raise ValueError(
                f"Relation #{i} ({parent_table}, {child_table}) occurs "
                f"{parents_and_children.count(relation[:2])} times. "
                f"Each relation must be unique."
            )

        # Check hierarchical keys
        _check_hierarchical_keys(
            i,
            parent_table,
            tables_spec[parent_table][1],
            child_table,
            tables_spec[child_table][1],
        )

    # Check there are no cycles
    _check_no_cycles(relations_spec, main_table_name)


def _check_hierarchical_keys(
    relation_id, parent_table, parent_table_key, child_table, child_table_key
):
    """Check that the parent table's key is contained in the child table's key"""
    # Perform the check and save the error status
    error_found = False
    if isinstance(parent_table_key, str) and isinstance(child_table_key, str):
        error_found = child_table_key != parent_table_key
    elif isinstance(parent_table_key, str) and is_list_like(child_table_key):
        error_found = parent_table_key not in child_table_key
    elif is_list_like(parent_table_key) and is_list_like(child_table_key):
        error_found = not set(parent_table_key).issubset(child_table_key)
    elif is_list_like(parent_table_key) and isinstance(child_table_key, str):
        error_found = (
            len(parent_table_key) != 1 or child_table_key not in parent_table_key
        )

    # Report any error found
    if error_found:
        if isinstance(child_table_key, str):
            child_table_key_msg = f"[{child_table_key}]"
        else:
            child_table_key_msg = f"[{', '.join(child_table_key)}]"
        if isinstance(parent_table_key, str):
            parent_table_key_msg = f"[{parent_table_key}]"
        else:
            parent_table_key_msg = f"[{', '.join(parent_table_key)}]"
        raise ValueError(
            f"Relation #{relation_id} child table '{child_table}' "
            f"key ({child_table_key_msg}) does not contain that of parent table "
            f"'{parent_table}' ({parent_table_key_msg})."
        )


def _check_no_cycles(relations_spec, main_table_name):
    """Check that there are no cycles in the 'relations' entry"""
    tables_to_visit = [main_table_name]
    tables_visited = set()
    while tables_to_visit:
        current_table = tables_to_visit.pop(0)
        tables_visited.add(current_table)
        for relation in relations_spec:
            parent_table, child_table = relation[:2]
            if parent_table == current_table:
                tables_to_visit.append(child_table)
                if tables_visited.intersection(tables_to_visit):
                    raise ValueError(
                        "'relations' entry contains a cycle that includes "
                        f"the relation ({parent_table}, {child_table})."
                    )


def _check_format_entry(format_spec):
    if not isinstance(format_spec, tuple):
        raise TypeError(type_error_message("'format' entry", format_spec, tuple))
    if len(format_spec) != 2:
        raise ValueError(
            f"'format' entry must be a tuple of size 2, not {len(format_spec)}"
        )
    sep, header = format_spec
    if not isinstance(sep, str):
        raise TypeError(
            type_error_message("'format' tuple's 1st element (separator)", sep, str)
        )
    if not isinstance(header, bool):
        raise TypeError(
            type_error_message("'format' tuple's 2nd element (header)", header, bool)
        )
    if len(sep) != 1:
        raise ValueError(f"'format' separator must be a single char, got '{sep}'")


def get_khiops_type(numpy_type):
    """Translates a numpy dtype to a Khiops dictionary type

    Parameters
    ----------
    numpy_type : `numpy.dtype`:
        Numpy type of the column

    Returns
    -------
    str
        Khiops type name. Either "Categorical", "Numerical" or "Timestamp"
    """
    lower_numpy_type = str(numpy_type).lower()

    # timedelta64 and datetime64 types
    if "datetime64" in lower_numpy_type or "timedelta64" in lower_numpy_type:
        khiops_type = "Timestamp"
    # float<x>, int<x>, uint<x> types
    elif "int" in lower_numpy_type or "float" in lower_numpy_type:
        khiops_type = "Numerical"
    # bool_ and object, character, bytes_, str_, void, record and other types
    else:
        khiops_type = "Categorical"

    return khiops_type


def get_khiops_variable_name(column_id):
    """Return the khiops variable name associated to a column id"""
    if isinstance(column_id, str):
        variable_name = column_id
    else:
        assert isinstance(column_id, np.int64)
        variable_name = f"Var{column_id}"
    return variable_name


def read_internal_data_table(file_path_or_stream, column_dtypes=None):
    """Reads into a DataFrame a data table file with the internal format settings

    The table is read with the following settings:

    - Use tab as separator
    - Read the column names from the first line
    - Use '"' as quote character
    - Use `csv.QUOTE_MINIMAL`
    - double quoting enabled (quotes within quotes can be escaped with '""')
    - UTF-8 encoding
    - User-specified dtypes (optional)

    Parameters
    ----------
    file_path_or_stream : str or file object
        The path of the internal data table file to be read or a readable file
        object.
    column_dtypes : dict, optional
        Dictionary linking column names with dtypes. See ``dtype`` parameter of the
        `pandas.read_csv` function. If not set, then the column types are detected
        automatically by pandas.

    Returns
    -------
    `pandas.DataFrame`
        The dataframe representation of the data table.
    """
    # Change the 'U' types (Unicode strings) to 'O' because pandas does not support them
    # in read_csv
    if column_dtypes is not None:
        execution_column_dtypes = {}
        for column_name, dtype in column_dtypes.items():
            if hasattr(dtype, "kind") and dtype.kind == "U":
                execution_column_dtypes[column_name] = np.dtype("O")
    else:
        execution_column_dtypes = None

    # Read and return the dataframe
    return pd.read_csv(
        file_path_or_stream,
        sep="\t",
        header=0,
        quotechar='"',
        quoting=csv.QUOTE_MINIMAL,
        doublequote=True,
        encoding="utf-8",
        dtype=execution_column_dtypes,
    )


def write_internal_data_table(dataframe, file_path_or_stream):
    """Writes a DataFrame to data table file with the internal format settings

    The table is written with the following settings:

    - Use tab as separator
    - Write the column names on the first line
    - Use '"' as quote character
    - Use `csv.QUOTE_MINIMAL`
    - double quoting enabled (quotes within quotes can be escaped with '""')
    - UTF-8 encoding
    - The index is not written

    Parameters
    ----------
    dataframe : `pandas.DataFrame`
        The dataframe to write.
    file_path_or_stream : str or file object
        The path of the internal data table file to be written or a writable file
        object.
    """
    dataframe.to_csv(
        file_path_or_stream,
        sep="\t",
        header=True,
        quotechar='"',
        quoting=csv.QUOTE_MINIMAL,
        doublequote=True,
        encoding="utf-8",
        index=False,
    )


def _column_or_1d_with_dtype(y, dtype=None):
    # 'dtype' has been introduced on `column_or_1d' since Scikit-learn 1.2;
    if sklearn.__version__ < "1.2":
        if pd.api.types.is_string_dtype(dtype) and y.isin(["True", "False"]).all():
            warnings.warn(
                "'y' stores strings restricted to 'True'/'False' values: "
                "The predict method may return a bool vector."
            )
        return column_or_1d(y, warn=True)
    else:
        return column_or_1d(y, warn=True, dtype=dtype)


class Dataset:
    """A representation of a dataset

    Parameters
    ----------
    X : `pandas.DataFrame` or dict (**Deprecated types**: tuple and list)
        Either:
          - A single dataframe
          - A ``dict`` dataset specification
    y : `pandas.Series` or str, optional
        The target column.
    categorical_target : bool, default True
        ``True`` if the vector ``y`` should be considered as a categorical variable. If
        ``False`` it is considered as numeric. Ignored if ``y`` is ``None``.
    key : str
        The name of the key column for all tables.
        **Deprecated:** Will be removed in khiops-python 11.
    """

    def __init__(self, X, y=None, categorical_target=True, key=None):
        # Initialize members
        self.main_table = None
        self.secondary_tables = None
        self.relations = None
        self.categorical_target = categorical_target
        self.target_column = None
        self.target_column_id = None
        self.sep = None
        self.header = None

        # Initialization from different types of input "X"
        # A single pandas dataframe
        if isinstance(X, pd.DataFrame):
            self.main_table = PandasTable("main_table", X)
            self.secondary_tables = []
        # A single numpy array (or compatible object)
        elif hasattr(X, "__array__"):
            self.main_table = NumpyTable("main_table", X)
            self.secondary_tables = []
        # A scipy.sparse.spmatrix
        elif isinstance(X, sp.spmatrix):
            self.main_table = SparseTable("main_table", X)
            self.secondary_tables = []
        # Special rejection for scipy.sparse.sparray (to pass the sklearn tests)
        # Note: We don't use scipy.sparse.sparray because it is not implemented in scipy
        # 1.10 which is the latest supporting py3.8
        elif isinstance(
            X,
            (
                sp.bsr_array,
                sp.coo_array,
                sp.csc_array,
                sp.csr_array,
                sp.dia_array,
                sp.dok_array,
                sp.lil_array,
            ),
        ):
            check_array(X, accept_sparse=False)
        # A tuple spec
        elif isinstance(X, tuple):
            warnings.warn(
                deprecation_message(
                    "Tuple dataset input",
                    "11.0.0",
                    replacement="dict dataset spec",
                    quote=False,
                ),
                stacklevel=3,
            )
            # Check the input tuple
            self._check_input_tuple(X)

            # Obtain path and separator
            path, sep = X

            # Initialization
            self.main_table = FileTable("main_table", path=path, sep=sep)
            self.secondary_tables = []

        # A dataset sequence spec
        # We try first for compatible python arrays then the deprecated sequences spec
        elif is_list_like(X):
            # Try to transform to a numerical array with sklearn's check_array
            # On failure we try the old deprecated sequence interface
            # When the old list interface is eliminated this will considerably reduce
            # this branch's code
            try:
                X_checked = check_array(X, ensure_2d=True, force_all_finite=False)
                self.main_table = NumpyTable("main_table", X_checked)
                self.secondary_tables = []
            except ValueError:
                warnings.warn(
                    deprecation_message(
                        "List dataset input",
                        "11.0.0",
                        replacement="dict dataset spec",
                        quote=False,
                    ),
                    stacklevel=3,
                )
                self._init_tables_from_sequence(X, key=key)
        # A a dataset dict spec
        elif is_dict_like(X):
            self._init_tables_from_mapping(X)
        # Fail if X is not recognized
        else:
            raise TypeError(
                type_error_message("X", X, "array-like", tuple, Sequence, Mapping)
            )

        # Initialization of the target column if any
        if y is not None:
            self._init_target_column(y)

        # Index the tables by name
        self._tables_by_name = {
            table.name: table for table in [self.main_table] + self.secondary_tables
        }

        # Deprecation warning for file-based datasets
        if isinstance(self.main_table, FileTable):
            warnings.warn(
                deprecation_message(
                    "File-based dataset spec",
                    "11.0.0",
                    "dataframe-based dataset or khiops.core API",
                    quote=False,
                ),
            )

        # Post-conditions
        assert self.main_table is not None, "'main_table' is 'None' after init"
        assert isinstance(
            self.secondary_tables, list
        ), "'secondary_tables' is not a list after init"
        assert not self.is_multitable or len(
            self.secondary_tables
        ), "'secondary_tables' is empty in a multi-table dataset"
        assert (
            y is None or self.target_column is not None
        ), "'y' is set but 'target_column' is None"

    def _check_input_tuple(self, X):
        if len(X) != 2:
            raise ValueError(f"'X' tuple input must have length 2 not {len(X)}")
        if not isinstance(X[0], str):
            raise TypeError(type_error_message("X[0]", X[0], str))
        if not isinstance(X[1], str):
            raise TypeError(type_error_message("X[1]", X[1], str))

    def _init_tables_from_sequence(self, X, key=None):
        """Initializes the spec from a list-like 'X'"""
        assert is_list_like(X), "'X' must be a list-like"

        # Check the input sequence
        self._check_input_sequence(X, key=key)

        # Initialize the tables
        if isinstance(X[0], pd.DataFrame):
            self.main_table = PandasTable("main_table", X[0], key=key)
            self.secondary_tables = []
            for index, dataframe in enumerate(X[1:], start=1):
                self.secondary_tables.append(
                    PandasTable(f"secondary_table_{index:02d}", dataframe, key=key)
                )
        else:
            self.main_table = FileTable("main_table", X[0], key=key)
            self.secondary_tables = []
            for index, table_path in enumerate(X[1:], start=1):
                self.secondary_tables.append(
                    FileTable(f"secondary_table_{index:02d}", table_path, key=key)
                )

        # Create a list of relations
        main_table_name = self.main_table.name
        self.relations = [
            (main_table_name, table.name, False) for table in self.secondary_tables
        ]

    def _check_input_sequence(self, X, key=None):
        # Check the first table
        if len(X) == 0:
            raise ValueError("'X' must be a non-empty sequence")
        if not isinstance(X[0], (str, pd.DataFrame)):
            raise TypeError(type_error_message("X[0]", X[0], str, pd.DataFrame))

        # Check that the secondary table types are coherent with that of the first
        main_table_type = type(X[0])
        for i, secondary_X in enumerate(X[1:], start=1):
            if not isinstance(secondary_X, main_table_type):
                raise TypeError(
                    type_error_message(f"Table at index {i}", X[i], main_table_type)
                    + " as the first table in X"
                )

        # Check the key for the main_table (it is the same for the others)
        _check_table_key("main_table", key)

    def _init_tables_from_mapping(self, X):
        """Initializes the table spec from a dict-like 'X'"""
        assert is_dict_like(X), "'X' must be dict-like"

        # Check the input mapping
        check_dataset_spec(X)

        # Initialize tables objects
        if len(X["tables"]) == 1:
            main_table_name = list(X["tables"])[0]
            main_table_source, main_table_key = list(X["tables"].values())[0]
            if isinstance(main_table_key, str):
                main_table_key = [main_table_key]
        else:
            main_table_name = X["main_table"]
            main_table_source, main_table_key = X["tables"][main_table_name]

        # Initialize a file dataset
        if isinstance(main_table_source, str):
            # Obtain the file format parameters
            if "format" in X:
                self.sep, self.header = X["format"]
            else:
                self.sep = "\t"
                self.header = True

            # Initialize the tables
            self.main_table = FileTable(
                main_table_name,
                main_table_source,
                key=main_table_key,
                sep=self.sep,
                header=self.header,
            )
            self.secondary_tables = []
            for table_name, (table_source, table_key) in X["tables"].items():
                if isinstance(table_key, str):
                    table_key = [table_key]
                if table_name != main_table_name:
                    self.secondary_tables.append(
                        FileTable(
                            table_name,
                            table_source,
                            key=table_key,
                            sep=self.sep,
                            header=self.header,
                        )
                    )
        # Initialize a Pandas dataset
        elif isinstance(main_table_source, pd.DataFrame):
            self.main_table = PandasTable(
                main_table_name,
                main_table_source,
                key=main_table_key,
            )
            self.secondary_tables = []
            for table_name, (table_source, table_key) in X["tables"].items():
                if table_name != main_table_name:
                    self.secondary_tables.append(
                        PandasTable(table_name, table_source, key=table_key)
                    )
        # Initialize a sparse dataset (monotable)
        elif isinstance(main_table_source, sp.spmatrix):
            self.main_table = SparseTable(
                main_table_name,
                main_table_source,
                key=main_table_key,
            )
            self.secondary_tables = []
        # Initialize a numpyarray dataset (monotable)
        else:
            self.main_table = NumpyTable(
                main_table_name,
                main_table_source,
            )
            if len(X["tables"]) > 1:
                raise ValueError(
                    "Multi-table schemas are only allowed "
                    "with pandas dataframe source tables"
                )
            self.secondary_tables = []

        # If the relations are not specified initialize to a star schema
        if "relations" not in X:
            self.relations = [
                (self.main_table.name, table.name, False)
                for table in self.secondary_tables
            ]
        # Otherwise initialize the relations in the spec
        else:
            relations = []
            for relation in X["relations"]:
                parent, child = relation[:2]
                relations.append(
                    (parent, child, relation[2] if len(relation) == 3 else False)
                )
            self.relations = relations

    def _init_target_column(self, y):
        assert self.main_table is not None
        assert self.secondary_tables is not None

        # Check y's type
        # For in memory target columns:
        # - column_or_1d checks *and transforms* to a numpy.array if successful
        # - warn=True in column_or_1d is necessary to pass sklearn checks
        if isinstance(y, str):
            y_checked = y
        else:
            if hasattr(y, "dtype"):
                if isinstance(y.dtype, pd.CategoricalDtype):
                    y_checked = _column_or_1d_with_dtype(
                        y, dtype=y.dtype.categories.dtype
                    )
                else:
                    y_checked = _column_or_1d_with_dtype(y, dtype=y.dtype)
            elif hasattr(y, "dtypes"):
                if isinstance(y.dtypes[0], pd.CategoricalDtype):
                    y_checked = _column_or_1d_with_dtype(
                        y, dtype=y.dtypes[0].categories.dtype
                    )
                else:
                    y_checked = _column_or_1d_with_dtype(y)
            else:
                y_checked = _column_or_1d_with_dtype(y)
        # Check the target type coherence with those of X's tables
        if isinstance(
            self.main_table, (PandasTable, SparseTable, NumpyTable)
        ) and isinstance(y_checked, str):
            if isinstance(self.main_table, PandasTable):
                type_message = "pandas.DataFrame"
            elif isinstance(self.main_table, SparseTable):
                type_message = "scipy.sparse.spmatrix"
            else:
                type_message = "numpy.ndarray"
            raise TypeError(
                type_error_message("y", y, "array-like")
                + f" (X's tables are of type {type_message})"
            )
        if isinstance(self.main_table.data_source, str) and not isinstance(
            y_checked, str
        ):
            raise TypeError(
                type_error_message("y", y, str)
                + " (X's tables are of type str [file paths])"
            )

        # Initialize the members related to the target
        # Case when y is a memory array
        if hasattr(y_checked, "__array__"):
            self.target_column = y_checked

            # Initialize the id of the target column
            if isinstance(y, pd.Series) and y.name is not None:
                self.target_column_id = y.name
            elif isinstance(y, pd.DataFrame):
                self.target_column_id = y.columns[0]
            else:
                if pd.api.types.is_integer_dtype(self.main_table.column_ids):
                    self.target_column_id = self.main_table.column_ids[-1] + 1
                else:
                    assert pd.api.types.is_string_dtype(self.main_table.column_ids)
                    self.target_column_id = "UnknownTargetColumn"

            # Fail if there is a column in the main_table with the target column's name
            if self.target_column_id in self.main_table.column_ids:
                raise ValueError(
                    f"Target column name '{self.target_column_id}' "
                    f"is already present in the main table. "
                    f"Column names: {list(self.main_table.column_ids)}"
                )
        # Case when y is column id: Set both the column and the id to it
        else:
            assert isinstance(y, str), type_error_message("y", y, str)
            self.target_column = y
            self.target_column_id = y

            # Check the target column exists in the main table
            if self.target_column_id not in self.main_table.column_ids:
                raise ValueError(
                    f"Target column '{self.target_column}' not present in main table. "
                    f"Column names: {list(self.main_table.column_ids)}'"
                )

            # Force the target column type from the parameters
            if self.categorical_target:
                self.main_table.khiops_types[self.target_column_id] = "Categorical"
            else:
                self.main_table.khiops_types[self.target_column_id] = "Numerical"

    @property
    def is_in_memory(self):
        """bool : ``True`` if the dataset is in-memory

        A dataset is in-memory if it is constituted either of only pandas.DataFrame
        tables, numpy.ndarray, or scipy.sparse.spmatrix tables.
        """

        return isinstance(self.main_table, (PandasTable, NumpyTable, SparseTable))

    @property
    def table_type(self):
        """type : The table type of this dataset's tables

        Possible values:

        - `PandasTable`
        - `NumpyTable`
        - `SparseTable`
        - `FileTable`
        """
        return type(self.main_table)

    @property
    def is_multitable(self):
        """bool : ``True`` if the dataset is multitable"""
        return self.secondary_tables is not None and len(self.secondary_tables) > 0

    def to_spec(self):
        """Returns a dictionary specification of this dataset"""
        ds_spec = {}
        ds_spec["main_table"] = self.main_table.name
        ds_spec["tables"] = {}
        ds_spec["tables"][self.main_table.name] = (
            self.main_table.data_source,
            self.main_table.key,
        )
        for table in self.secondary_tables:
            ds_spec["tables"][table.name] = (table.data_source, table.key)
        if self.relations:
            ds_spec["relations"] = []
            ds_spec["relations"].extend(self.relations)
        if self.table_type == FileTable:
            ds_spec["format"] = (self.sep, self.header)

        return ds_spec

    def copy(self):
        """Creates a copy of the dataset

        Referenced pandas.DataFrame's, numpy.nparray's and scipy.sparse.spmatrix's in
        tables are copied as references.
        """
        return Dataset(self.to_spec())

    def get_table(self, table_name):
        """Returns a table by its name

        Parameters
        ----------
        table_name: str
            The name of the table to be retrieved.

        Returns
        -------
        `DatasetTable`
            The table object for the specified name.

        Raises
        ------
        `KeyError`
            If there is no table with the specified name.
        """
        return self._tables_by_name[table_name]

    def create_khiops_dictionary_domain(self):
        """Creates a Khiops dictionary domain representing this dataset

        Returns
        -------
        `.DictionaryDomain`
            The dictionary domain object representing this dataset
        """
        assert self.main_table is not None, "'main_table' must be initialized"

        # Create root dictionary and add it to the domain
        dictionary_domain = kh.DictionaryDomain()
        main_dictionary = self.main_table.create_khiops_dictionary()
        dictionary_domain.add_dictionary(main_dictionary)

        # For in-memory datasets: Add the target variable if available
        if self.is_in_memory and self.target_column is not None:
            variable = kh.Variable()
            variable.name = get_khiops_variable_name(self.target_column_id)
            if self.categorical_target:
                variable.type = "Categorical"
            else:
                variable.type = "Numerical"
            main_dictionary.add_variable(variable)

        # Create the dictionaries for each secondary table and the table variables in
        # root dictionary that point to each secondary table
        # This is performed using a breadth-first-search over the graph of relations
        # Note: In general 'name' and 'object_type' fields of Variable can be different
        if self.secondary_tables:
            main_dictionary.root = True
            table_names = [table.name for table in self.secondary_tables]
            tables_to_visit = [self.main_table.name]
            while tables_to_visit:
                current_table = tables_to_visit.pop(0)
                for relation in self.relations:
                    parent_table, child_table, is_one_to_one_relation = relation
                    if parent_table == current_table:
                        tables_to_visit.append(child_table)
                        parent_table_name = parent_table
                        index_table = table_names.index(child_table)
                        table = self.secondary_tables[index_table]
                        parent_table_dictionary = dictionary_domain.get_dictionary(
                            parent_table_name
                        )
                        dictionary = table.create_khiops_dictionary()
                        dictionary_domain.add_dictionary(dictionary)
                        table_variable = kh.Variable()
                        if is_one_to_one_relation:
                            table_variable.type = "Entity"
                        else:
                            table_variable.type = "Table"
                        table_variable.name = table.name
                        table_variable.object_type = table.name
                        parent_table_dictionary.add_variable(table_variable)

        return dictionary_domain

    def create_table_files_for_khiops(self, output_dir, sort=True):
        """Prepares the tables of the dataset to be used by Khiops

        If this is a multi-table dataset it will create sorted copies the tables.

        Parameters
        ----------
        output_dir : str
            The directory where the sorted tables will be created.

        Returns
        -------
        tuple
            A tuple containing:

            - The path of the main table
            - A dictionary containing the relation [table-name -> file-path] for the
              secondary tables. The dictionary is empty for monotable datasets.
        """
        # Sort the main table unless:
        # - The caller specifies not to do it (sort = False)
        # - The dataset is mono-table and the main table has no key
        sort_main_table = sort and (
            self.is_multitable or self.main_table.key is not None
        )

        # In-memory dataset: Create the table files and add the target column
        if self.is_in_memory:
            main_table_path = self.main_table.create_table_file_for_khiops(
                output_dir,
                sort=sort_main_table,
                target_column=self.target_column,
                target_column_id=self.target_column_id,
            )
        # File dataset: Create the table files (the target column is in the file)
        else:
            main_table_path = self.main_table.create_table_file_for_khiops(
                output_dir,
                sort=sort_main_table,
            )

        # Create a copy of each secondary table
        secondary_table_paths = {}
        for table in self.secondary_tables:
            secondary_table_paths[table.name] = table.create_table_file_for_khiops(
                output_dir, sort=sort
            )

        return main_table_path, secondary_table_paths

    def __repr__(self):
        return str(self.create_khiops_dictionary_domain())


# pylint: enable=invalid-name


class DatasetTable(ABC):
    """A generic dataset table"""

    def __init__(self, name, key=None):
        # Check input
        if not isinstance(name, str):
            raise TypeError(type_error_message("name", name, str))
        if not name:
            raise ValueError("'name' cannot be empty")
        if key is not None:
            if not is_list_like(key) and not isinstance(key, (str, int)):
                raise TypeError(type_error_message("key", key, str, int, "list-like"))
            if is_list_like(key):
                for column_index, column_id in enumerate(key):
                    if not isinstance(column_id, (str, int)):
                        raise TypeError(
                            type_error_message(
                                f"key[{column_index}]", column_id, str, int
                            )
                            + f" at table '{name}'"
                        )

        # Initialization (must be completed by concrete sub-classes)
        self.name = name
        self.data_source = None
        if is_list_like(key) or key is None:
            self.key = key
        else:
            self.key = [key]
        self.column_ids = None
        self.khiops_types = None
        self.n_samples = None

    def check_key(self):
        """Checks that the key columns exist"""
        if self.key is not None:
            if not is_list_like(self.key):
                raise TypeError(
                    type_error_message("key", self.key, str, int, "list-like")
                )
            for column_name in self.key:
                if column_name not in self.column_ids:
                    raise ValueError(
                        f"Column '{column_name}' not present in table '{self.name}'"
                    )

    @abstractmethod
    def create_table_file_for_khiops(self, output_dir, sort=True):
        """Creates a copy of the table at the specified directory"""

    def n_features(self):
        """Returns the number of features of the table

        The target column does not count.
        """
        return len(self.column_ids)

    def create_khiops_dictionary(self):
        """Creates a Khiops dictionary representing this table

        Returns
        -------
        `.Dictionary`:
            The Khiops Dictionary object describing this table's schema

        """
        assert self.column_ids is not None, "Dataset column list is None"
        assert self.key is None or is_list_like(self.key), "'key' is not list-like"

        # Create dictionary object
        dictionary = kh.Dictionary()
        dictionary.name = self.name
        if self.key is not None:
            dictionary.key = self.key

        # For each column add a Khiops variable to the dictionary
        for column_id in self.column_ids:
            variable = kh.Variable()
            variable.name = get_khiops_variable_name(column_id)

            # Set the type of the column/variable
            # Case of a column in the key : Set to categorical
            if self.key is not None and column_id in self.key:
                variable.type = "Categorical"
            # The rest of columns: Obtain the type from dtypes
            else:
                variable.type = self.khiops_types[column_id]
            dictionary.add_variable(variable)
        return dictionary


class PandasTable(DatasetTable):
    """DatasetTable encapsulating a pandas dataframe

    Parameters
    ----------
    name : str
        Name for the table.
    dataframe : `pandas.DataFrame`
        The data frame to be encapsulated. It must be non-empty.
    key : list of str, optional
        The names of the columns composing the key.
    """

    def __init__(self, name, dataframe, key=None):
        # Call the parent method
        super().__init__(name, key=key)

        # Check inputs specific to this sub-class
        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError(type_error_message("dataframe", dataframe, pd.DataFrame))
        if dataframe.shape[0] == 0:
            raise ValueError("'dataframe' is empty")

        # Initialize the attributes
        self.data_source = dataframe
        self.n_samples = len(self.data_source)

        # Initialize feature columns and verify their types
        self.column_ids = self.data_source.columns.values
        if not np.issubdtype(self.column_ids.dtype, np.integer):
            if np.issubdtype(self.column_ids.dtype, object):
                for i, column_id in enumerate(self.column_ids):
                    if not isinstance(column_id, str):
                        raise TypeError(
                            f"Dataframe column ids must be either all integers or "
                            f"all strings. Column id at index {i} ('{column_id}') is"
                            f" of type '{type(column_id).__name__}'"
                        )
            else:
                raise TypeError(
                    f"Dataframe column ids must be either all integers or "
                    f"all strings. The column index has dtype "
                    f"'{self.column_ids.dtype}'"
                )

        # Initialize Khiops types
        self.khiops_types = {
            column_id: get_khiops_type(self.data_source.dtypes[column_id])
            for column_id in self.column_ids
        }

        # Check key integrity
        self.check_key()

    def __repr__(self):
        dtypes_str = (
            str(self.data_source.dtypes).replace("\n", ", ")[:-16].replace("    ", ":")
        )
        return (
            f"<{self.__class__.__name__}; cols={list(self.column_ids)}; "
            f"dtypes={dtypes_str}>"
        )

    def get_column_dtype(self, column_id):
        if column_id not in self.data_source.dtypes:
            raise KeyError(f"Column '{column_id}' not found in the dtypes field")
        return self.data_source.dtypes[column_id]

    def create_table_file_for_khiops(
        self, output_dir, sort=True, target_column=None, target_column_id=None
    ):
        assert not sort or self.key is not None, "Cannot sort table without a key"
        assert not sort or is_list_like(
            self.key
        ), "Cannot sort table with a key is that is not list-like"
        assert not sort or len(self.key) > 0, "Cannot sort table with an empty key"
        assert target_column is not None or target_column_id is None
        assert target_column_id is not None or target_column is None

        # Create the output table resource object
        output_table_path = fs.get_child_path(output_dir, f"{self.name}.txt")

        # Write the output dataframe
        output_dataframe = self.data_source.copy()
        output_names = {
            column_id: get_khiops_variable_name(column_id)
            for column_id in self.column_ids
        }
        output_dataframe.rename(columns=output_names, inplace=True)
        if target_column is not None:
            output_dataframe[get_khiops_variable_name(target_column_id)] = (
                target_column.copy()
            )

        # Sort by key if requested (as string)
        if sort:
            output_dataframe.sort_values(
                by=self.key,
                key=lambda array: array.astype("str"),
                inplace=True,
                kind="mergesort",
            )

        # Write the dataframe to an internal table file
        with io.StringIO() as output_dataframe_stream:
            write_internal_data_table(output_dataframe, output_dataframe_stream)
            fs.write(
                output_table_path, output_dataframe_stream.getvalue().encode("utf-8")
            )

        return output_table_path


class NumpyTable(DatasetTable):
    """DatasetTable encapsulating a NumPy array

    Parameters
    ----------
    name : str
        Name for the table.
    array : `numpy.ndarray` of shape (n_samples, n_features_in)
        The data frame to be encapsulated.
    key : :external:term`array-like` of int, optional
        The names of the columns composing the key.
    """

    def __init__(self, name, array, key=None):
        # Call the parent method
        super().__init__(name, key=key)

        # Check the array's types and shape
        if not hasattr(array, "__array__"):
            raise TypeError(type_error_message("array", array, np.ndarray))

        # Initialize the members
        self.data_source = check_array(array, ensure_2d=True, force_all_finite=False)
        self.column_ids = column_or_1d(range(self.data_source.shape[1]))
        self.khiops_types = {
            column_id: get_khiops_type(self.data_source.dtype)
            for column_id in self.column_ids
        }
        self.n_samples = len(self.data_source)

    def __repr__(self):
        dtype_str = str(self.data_source.dtype)
        return (
            f"<{self.__class__.__name__}; cols={list(self.column_ids)}; "
            f"dtype={dtype_str}; target={self.target_column_id}>"
        )

    def get_column_dtype(self, _):
        return self.data_source.dtype

    def create_table_file_for_khiops(
        self, output_dir, sort=True, target_column=None, target_column_id=None
    ):
        assert not sort or self.key is not None, "Cannot sort table without a key"
        assert not sort or is_list_like(
            self.key
        ), "Cannot sort table with a key is that is not list-like"
        assert not sort or len(self.key) > 0, "Cannot sort table with an empty key"

        # Create the output table resource object
        output_table_path = fs.get_child_path(output_dir, f"{self.name}.txt")

        # Write the output dataframe
        # Note: This is not optimized for memory.
        output_dataframe = pd.DataFrame(self.data_source.copy())
        output_dataframe.columns = [
            get_khiops_variable_name(column_id) for column_id in self.column_ids
        ]
        if target_column is not None:
            output_dataframe[get_khiops_variable_name(target_column_id)] = (
                target_column.copy()
            )

        # Sort by key if requested (as string)
        if sort:
            np.sort(
                output_dataframe,
                by=self.key,
                key=lambda array: array.astype("str"),
                inplace=True,
                kind="mergesort",
            )

        # Write the dataframe to an internal table file
        with io.StringIO() as output_dataframe_stream:
            write_internal_data_table(output_dataframe, output_dataframe_stream)
            fs.write(
                output_table_path, output_dataframe_stream.getvalue().encode("utf-8")
            )

        return output_table_path


class SparseTable(DatasetTable):
    """DatasetTable encapsulating a SciPy sparse matrix

    Parameters
    ----------
    name : str
        Name for the table.
    matrix : `scipy.sparse.spmatrix`
        The sparse matrix to be encapsulated.
    key : list of str, optional
        The names of the columns composing the key.
    """

    def __init__(self, name, matrix, key=None):
        assert key is None, "'key' must be unset for sparse matrix tables"
        # Call the parent method
        super().__init__(name, key=key)

        # Check the sparse matrix types
        if not isinstance(matrix, sp.spmatrix):
            raise TypeError(
                type_error_message("matrix", matrix, "scipy.sparse.spmatrix")
            )
        if not np.issubdtype(matrix.dtype, np.number):
            raise TypeError(
                type_error_message("'matrix' dtype", matrix.dtype, "numeric")
            )

        # Initialize the members
        self.data_source = matrix
        self.column_ids = column_or_1d(range(matrix.shape[1]))
        self.khiops_types = {
            column_id: get_khiops_type(matrix.dtype) for column_id in self.column_ids
        }
        self.n_samples = self.data_source.shape[0]

    def __repr__(self):
        dtype_str = str(self.data_source.dtype)
        return (
            f"<{self.__class__.__name__}; cols={list(self.column_ids)}; "
            f"dtype={dtype_str}>"
        )

    def get_column_dtype(self, _):
        return self.data_source.dtype

    def create_khiops_dictionary(self):
        """Creates a Khiops dictionary representing this sparse table

        Adds metadata to each sparse variable

        Returns
        -------
        `.Dictionary`:
            The Khiops Dictionary object describing this table's schema

        """

        # create dictionary as usual
        dictionary = super().create_khiops_dictionary()

        # create variable block for containing the sparse variables
        variable_block = VariableBlock()
        variable_block.name = "SparseVariables"

        # For each variable, add metadata, named `VarKey`
        variable_names = [variable.name for variable in dictionary.variables]
        for i, variable_name in enumerate(variable_names, 1):
            variable = dictionary.remove_variable(variable_name)
            variable.meta_data.add_value("VarKey", i)
            variable_block.add_variable(variable)
        dictionary.add_variable_block(variable_block)

        return dictionary

    def _flatten(self, iterable):
        if isinstance(iterable, Iterable):
            for iterand in iterable:
                if isinstance(iterand, Iterable):
                    yield from self._flatten(iterand)
                else:
                    yield iterand

    def _write_sparse_block(self, row_index, stream, target_value=None):

        # Access the sparse row
        row = self.data_source.getrow(row_index)
        # Variable indices are not always sorted in `row.indices`
        # Khiops needs variable indices to be sorted
        sorted_indices = np.sort(row.nonzero()[1], axis=-1, kind="mergesort")

        # Flatten row for Python < 3.9 scipy.sparse.lil_matrix whose API
        # is not homogeneous with other sparse matrices: it stores
        # opaque Python lists as elements
        # Thus:
        # - if isinstance(self.data_source, sp.lil_matrix) and Python 3.8, then
        # row.data is np.array([list([...])])
        # - else, row.data is np.array([...])
        # TODO: remove this flattening once Python 3.8 support is dropped
        sorted_data = np.fromiter(self._flatten(row.data), row.data.dtype)[
            sorted_indices.argsort()
        ]
        for variable_index, variable_value in zip(sorted_indices, sorted_data):
            stream.write(f"{variable_index + 1}:{variable_value} ")

        # Write the target value at the end of the record if available
        if target_value is not None:
            stream.write(f"\t{target_value}\n")
        else:
            stream.write("\n")

    def create_table_file_for_khiops(
        self, output_dir, sort=True, target_column=None, target_column_id=None
    ):
        assert target_column is not None or target_column_id is None
        assert target_column_id is not None or target_column is None

        # Create the output table resource object
        output_table_path = fs.get_child_path(output_dir, f"{self.name}.txt")

        # Write the sparse matrix to an internal table file
        with io.StringIO() as output_sparse_matrix_stream:
            if target_column is not None:
                output_sparse_matrix_stream.write(
                    f"SparseVariables\t{get_khiops_variable_name(target_column_id)}\n"
                )
                for target_value, row_index in zip(
                    target_column, range(self.data_source.shape[0])
                ):
                    self._write_sparse_block(
                        row_index,
                        output_sparse_matrix_stream,
                        target_value=target_value,
                    )
            else:
                output_sparse_matrix_stream.write("SparseVariables\n")
                for row_index in range(self.data_source.shape[0]):
                    self._write_sparse_block(row_index, output_sparse_matrix_stream)
            fs.write(
                output_table_path,
                output_sparse_matrix_stream.getvalue().encode("utf-8"),
            )

        return output_table_path


class FileTable(DatasetTable):
    """DatasetTable encapsulating a delimited text data file

    Parameters
    ----------
    name : str
        Name for the table.
    path : str
        Path of the file containing the table.
    key : list-like of str, optional
        The names of the columns composing the key.
    sep : str, optional
        Field separator character. If not specified it will be inferred from the file.
    header : bool, optional
        Indicates if the table.
    """

    def __init__(
        self,
        name,
        path,
        key=None,
        sep="\t",
        header=True,
    ):
        # Initialize parameters
        super().__init__(name=name, key=key)

        # Check the parameters specific to this sub-class
        if not isinstance(path, str):
            raise TypeError(type_error_message("path", path, str))
        if not fs.exists(path):
            raise ValueError(f"Non-existent data table file: {path}")

        # Initialize members specific to this sub-class
        self.data_source = path
        self.sep = sep
        self.header = header

        # Build a dictionary file from the input data table
        # Note: We use export_dictionary_as_json instead of read_dictionary_file
        #       because it makes fail the sklearn mocked tests (this is technical debt)
        try:
            tmp_kdic_path = kh.get_runner().create_temp_file("file_table_", ".kdic")
            tmp_kdicj_path = kh.get_runner().create_temp_file("file_table_", ".kdicj")
            kh.build_dictionary_from_data_table(
                self.data_source,
                self.name,
                tmp_kdic_path,
                field_separator=self.sep,
                header_line=header,
            )
            kh.export_dictionary_as_json(tmp_kdic_path, tmp_kdicj_path)
            json_domain = json.loads(fs.read(tmp_kdicj_path))
        finally:
            fs.remove(tmp_kdic_path)
            fs.remove(tmp_kdicj_path)

        # Alert the user if the parsing failed
        if len(json_domain["dictionaries"]) == 0:
            raise KhiopsRuntimeError(
                f"Failed to build a dictionary "
                f"from data table file: {self.data_source}"
            )

        # Set the column names and types
        variables = json_domain["dictionaries"][0]["variables"]
        self.column_ids = [var["name"] for var in variables]
        self.khiops_types = {var["name"]: var["type"] for var in variables}

        # Check key integrity
        self.check_key()

    def create_table_file_for_khiops(self, output_dir, sort=True):
        assert not sort or self.key is not None, "key is 'None'"

        # Create the input and output file resources
        if sort:
            output_table_file_path = fs.get_child_path(
                output_dir, f"sorted_{self.name}.txt"
            )
        else:
            output_table_file_path = fs.get_child_path(
                output_dir, f"copy_{self.name}.txt"
            )

        # Fail if they have the same path
        if output_table_file_path == self.data_source:
            raise ValueError(f"Cannot overwrite this table's path: {self.data_source}")

        # Create a sorted copy if requested
        if sort:
            # Create the sorting dictionary domain
            sort_dictionary_domain = kh.DictionaryDomain()
            sort_dictionary_domain.add_dictionary(self.create_khiops_dictionary())

            # Delegate the sorting and copy to khiops.core.sort_data_table
            # We use the same input format of the original table
            kh.sort_data_table(
                sort_dictionary_domain,
                self.name,
                self.data_source,
                output_table_file_path,
                self.key,
                field_separator=self.sep,
                header_line=self.header,
                output_field_separator=self.sep,
                output_header_line=self.header,
            )

        # Otherwise copy the contents to the output file
        else:
            fs.write(output_table_file_path, fs.read(self.data_source))

        return output_table_file_path
